Type: Package
Package: systemfonts
Title: System Native Font Finding
Version: 1.3.2
Authors@R: c(
    person("Thomas Lin", "Pedersen", , "thomas.pedersen@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-5147-4711")),
    person("Jeroen", "Ooms", , "jeroen@berkeley.edu", role = "aut",
           comment = c(ORCID = "0000-0002-4035-0289")),
    person("Devon", "Govett", role = "aut",
           comment = "Author of font-manager"),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Provides system native access to the font catalogue. As font
    handling varies between systems it is difficult to correctly locate
    installed fonts across different operating systems. The 'systemfonts'
    package provides bindings to the native libraries on Windows, macOS
    and Linux for finding font files that can then be used further by e.g.
    graphic devices. The main use is intended to be from compiled code but
    'systemfonts' also provides access from R.
License: MIT + file LICENSE
URL: https://github.com/r-lib/systemfonts,
        https://systemfonts.r-lib.org
BugReports: https://github.com/r-lib/systemfonts/issues
Depends: R (>= 3.2.0)
Imports: base64enc, grid, jsonlite, lifecycle, tools, utils
Suggests: covr, farver, ggplot2, graphics, knitr, ragg, rmarkdown,
        svglite, testthat (>= 2.1.0)
LinkingTo: cpp11 (>= 0.2.1)
VignetteBuilder: knitr
Config/build/compilation-database: true
Config/Needs/website: tidyverse/tidytemplate
Config/usethis/last-upkeep: 2025-04-23
Encoding: UTF-8
RoxygenNote: 7.3.2
SystemRequirements: fontconfig, freetype2
NeedsCompilation: yes
Packaged: 2026-03-05 12:24:52 UTC; thomas
Author: Thomas Lin Pedersen [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-5147-4711>),
  Jeroen Ooms [aut] (ORCID: <https://orcid.org/0000-0002-4035-0289>),
  Devon Govett [aut] (Author of font-manager),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Thomas Lin Pedersen <thomas.pedersen@posit.co>
Repository: CRAN
Date/Publication: 2026-03-05 13:10:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 23:16:44 UTC; unix
Archs: systemfonts.so.dSYM

Type: Package
Package: systemfonts
Title: System Native Font Finding
Version: 1.3.2
Authors@R: c(
    person("Thomas Lin", "Pedersen", , "thomas.pedersen@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-5147-4711")),
    person("Jeroen", "Ooms", , "jeroen@berkeley.edu", role = "aut",
           comment = c(ORCID = "0000-0002-4035-0289")),
    person("Devon", "Govett", role = "aut",
           comment = "Author of font-manager"),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Provides system native access to the font catalogue. As font
    handling varies between systems it is difficult to correctly locate
    installed fonts across different operating systems. The 'systemfonts'
    package provides bindings to the native libraries on Windows, macOS
    and Linux for finding font files that can then be used further by e.g.
    graphic devices. The main use is intended to be from compiled code but
    'systemfonts' also provides access from R.
License: MIT + file LICENSE
URL: https://github.com/r-lib/systemfonts,
        https://systemfonts.r-lib.org
BugReports: https://github.com/r-lib/systemfonts/issues
Depends: R (>= 3.2.0)
Imports: base64enc, grid, jsonlite, lifecycle, tools, utils
Suggests: covr, farver, ggplot2, graphics, knitr, ragg, rmarkdown,
        svglite, testthat (>= 2.1.0)
LinkingTo: cpp11 (>= 0.2.1)
VignetteBuilder: knitr
Config/build/compilation-database: true
Config/Needs/website: tidyverse/tidytemplate
Config/usethis/last-upkeep: 2025-04-23
Encoding: UTF-8
RoxygenNote: 7.3.2
SystemRequirements: fontconfig, freetype2
NeedsCompilation: yes
Packaged: 2026-03-05 12:24:52 UTC; thomas
Author: Thomas Lin Pedersen [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-5147-4711>),
  Jeroen Ooms [aut] (ORCID: <https://orcid.org/0000-0002-4035-0289>),
  Devon Govett [aut] (Author of font-manager),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Thomas Lin Pedersen <thomas.pedersen@posit.co>
Repository: CRAN
Date/Publication: 2026-03-05 13:10:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-27 05:41:30 UTC; unix
Archs: systemfonts.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("$",font_variation)
S3method("$<-",font_variation)
S3method("[",font_variation)
S3method("[<-",font_variation)
S3method("[[",font_variation)
S3method("[[<-",font_variation)
S3method(c,font_feature)
S3method(format,font_feature)
S3method(format,font_variation)
S3method(length,font_feature)
S3method(length,font_variation)
S3method(names,font_variation)
S3method(print,font_feature)
S3method(print,font_variation)
export(add_fonts)
export(as_font_weight)
export(as_font_width)
export(clear_local_fonts)
export(clear_registry)
export(font_fallback)
export(font_feature)
export(font_info)
export(font_variation)
export(fonts_as_import)
export(get_fallback)
export(get_from_font_squirrel)
export(get_from_google_fonts)
export(glyph_info)
export(glyph_outline)
export(glyph_raster)
export(glyph_raster_grob)
export(match_font)
export(match_fonts)
export(plot_glyph_stats)
export(register_font)
export(register_variant)
export(registry_fonts)
export(require_font)
export(reset_font_cache)
export(scan_local_fonts)
export(search_web_fonts)
export(shape_string)
export(str_split_emoji)
export(string_metrics_dev)
export(string_width)
export(string_widths_dev)
export(system_fonts)
importFrom(lifecycle,deprecated)
useDynLib(systemfonts, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# systemfonts <a href="https://systemfonts.r-lib.org"><img src="man/figures/logo.png" align="right" height="138" alt="systemfonts website" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/systemfonts)](https://cran.r-project.org/package=systemfonts)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![R-CMD-check](https://github.com/r-lib/systemfonts/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/systemfonts/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/systemfonts/graph/badge.svg)](https://app.codecov.io/gh/r-lib/systemfonts)
<!-- badges: end -->

systemfonts is a package that locates installed fonts. It uses the
system-native libraries on Mac (CoreText) and Linux (FontConfig), and
uses Freetype to parse the fonts in the registry on Windows.

## Installation

systemfonts is available from CRAN using
`install.packages('systemfonts')`. It is however still under development
and you can install the development version using devtools.

``` r
# install.packages('pak')
pak::pak('r-lib/systemfonts')
```

## Examples

The main use of this package is to locate font files based on family and
style:

``` r
library(systemfonts)

match_fonts('Avenir', italic = TRUE)
#>                               path index features variations
#> 1 /System/Library/Fonts/Avenir.ttc     1
```

This function returns the path to the file holding the font, as well as
the 0-based index of the font in the file.

It is also possible to get a data.frame of all available fonts:

``` r
system_fonts()
#> # A tibble: 1,221 × 10
#>    path          index name  family style weight width italic monospace variable
#>    <chr>         <int> <chr> <chr>  <chr> <ord>  <ord> <lgl>  <lgl>     <lgl>   
#>  1 /Users/thoma…     0 Recu… Recur… Sans… medium norm… FALSE  FALSE     TRUE    
#>  2 /Users/thoma…     0 Syne… Syne   Regu… normal norm… FALSE  FALSE     FALSE   
#>  3 /Users/thoma…     0 Inpu… Input… Extr… light  cond… TRUE   TRUE      FALSE   
#>  4 /System/Libr…     1 Telu… Telug… Bold  bold   norm… FALSE  FALSE     FALSE   
#>  5 /Users/thoma…     0 Inpu… Input… Thin  light  semi… FALSE  FALSE     FALSE   
#>  6 /System/Libr…     0 Micr… Micro… Regu… normal norm… FALSE  FALSE     FALSE   
#>  7 /Users/thoma…     0 Fira… Fira … Bold  bold   norm… FALSE  TRUE      FALSE   
#>  8 /System/Libr…     0 Sukh… Sukhu… Thin  ultra… semi… FALSE  FALSE     FALSE   
#>  9 /System/Libr…     0 Sathu Sathu  Regu… normal norm… FALSE  FALSE     FALSE   
#> 10 /Users/thoma…     0 IBMP… IBM P… Text… normal norm… TRUE   FALSE     FALSE   
#> # ℹ 1,211 more rows
```

Further, you can query additional information about fonts and specific
glyphs, if that is of interest using the `font_info()` and
`glyph_info()` functions.

### Custom fonts

While the package was created to provide transparent access to fonts
installed on the system, it has grown to also provide ways to work with
font files not part of the system installation. This is especially
beneficial if you are running code on a system where you don’t have
administrator rights and need to use a custom font.

systemfonts provide the `add_fonts()` function which takes a vector of
file paths and add these to the lookup database without installing them.
Further, systemfonts automatically looks in the `./fonts` and `~/fonts`
folders and adds any font files located there during startup. This means
that you can distribute a script along with a fonts directory and have
those fonts automatically available during execution of the script.

In addition to the above, systemfonts also provides access to online
font repositories such as [Google Fonts](https://fonts.google.com) and
can search and download from these, automatically adding the downloaded
fonts to the lookup database.

All these functionalities are condensed into a single function,
`require_font()`, which allows you to state a font dependency inside a
script. The function will first look for the font on the system, and
failing that, will try to fetch it from an online repository. If that
fails it will either throw an error or remap the font to another of the
developers choosing.

## C API

While getting this information in R is nice, the intended use is mostly
through compiled code so that graphic devices can easily locate relevant
font files etc.

In order to use functions from systemfonts in C(++) code your package
should list systemfonts in the `LinkingTo` field in the `DESCRIPTION`
file. Once this is done you can now `#include <systemfonts.h>` in your
code and use the provided functions. Look into the
[`inst/include/systemfonts.h`](https://github.com/r-lib/systemfonts/blob/master/inst/include/systemfonts.h)
file to familiarise yourself with the C API.

## System Defaults

systemfonts will always try to find a font for you, even if none exist
with the given family name or style. How it resolves this is system
specific and should not be relied on, but it can be expected that a
valid font file is always returned no matter the input.

A few special aliases exist that behaves predictably but system
dependent:

- `""` and `"sans"` return *Helvetica* on Mac, *Arial* on Windows, and
  the default sans-serif font on Linux (*DejaVu Sans* on Ubuntu)
- `"serif"` return *Times* on Mac, *Times New Roman* on Windows, and the
  default serif font on Linux (*DejaVu Serif* on Ubuntu)
- `"mono"` return *Courier* on Mac, *Courier New* on Windows, and the
  default mono font on Linux (*DejaVu Mono* on Ubuntu)
- `"emoji"` return *Apple Color Emoji* on Mac, *Segoe UI Emoji* on
  Windows, and the default emoji font on Linux (*Noto Color* on Ubuntu)

## Code of Conduct

Please note that the ‘systemfonts’ project is released with a
[Contributor Code of
Conduct](https://github.com/r-lib/systemfonts/blob/master/CODE_OF_CONDUCT.md).
By contributing to this project, you agree to abide by its terms.

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
  dev = "ragg_png",
  dpi = 144,
  collapse = TRUE,
  comment = "#>",
  fig.asp = NULL,
  fig.height = 4.326,
  fig.width = 7,
  eval = FALSE
)

## -----------------------------------------------------------------------------
# st <- marquee::classic_style(30, body_font = "Avenir Next", lineheight = 1, margin = marquee::trbl(4)) |>
#   marquee::modify_style("anul", weight = "thin") |>
#   marquee::modify_style("anub", weight = "ultrabold")
# text <- paste(
#   "{.anul Avenir Next Ultra Light}  ",
#   "{.anul *Avenir Next Ultra Light Italic*}  ",
#   "Avenir Next  ",
#   "*Avenir Next Italic*  ",
#   "{.anub Avenir Next Ultra Bold}  ",
#   "{.anub *Avenir Next Ultra Bold Italic*}  ",
#   sep = "\n"
# )
# grid::grid.draw(
#   marquee::marquee_grob(text, st)
# )

## -----------------------------------------------------------------------------
# systemfonts::plot_glyph_stats("j", family = "Helvetica", size = 30)

#' Find a system font by name and style
#'
#' This function locates the font file (and index) best matching a name and
#' optional style (italic/bold). A font file will be returned even if a match
#' isn't found, but it is not necessarily similar to the requested family and
#' it should not be relied on for font substitution. The aliases `"sans"`,
#' `"serif"`, and `"mono"` match to the system default sans-serif, serif, and
#' mono fonts respectively (`""` is equivalent to `"sans"`).
#'
#'
#' @param family The name of the font family
#' @param italic,bold logicals indicating the font style
#'
#' @return A list containing the path locating the font file and the 0-based
#' index of the font in the file.
#'
#' @export
#'
#' @examples
#' # Get the system default sans-serif font in italic
#' match_font('sans', italic = TRUE)
#'
match_font <- function(family, italic = FALSE, bold = FALSE) {
  if (!is.character(family)) stop("family must be a string", call. = FALSE)
  match_font_c(family, as.logical(italic), as.logical(bold))
}

context("Font Matching")

sysname <- tolower(Sys.info()[["sysname"]])
font <- switch(
  sysname,
  darwin = "Helvetica",
  linux = "DejaVuSans",
  windows = "arial"
)

test_that("Font files can be found", {
  font_path <- match_font("sans")$path

  expect_true(file.exists(font_path))

  skip_on_os("linux") # Different fonts for different distros
  skip_on_os("solaris") # Have no idea what it is supposed to give
  expect_equal(tools::file_path_sans_ext(basename(font_path)), font)
})

test_that("Default font is correct", {
  font_path <- match_font("sans")$path

  expect_true(file.exists(font_path))

  skip_on_os("linux")
  skip_on_os("solaris")
  expect_equal(tools::file_path_sans_ext(basename(font_path)), font)
})

#' Split a string into emoji and non-emoji glyph runs
#' 
#' In order to do correct text rendering, the font needed must be figured out. A
#' common case is rendering of emojis within a string where the system emoji 
#' font is used rather than the requested font. This function will inspect the
#' provided strings and split them up in runs that must be rendered with the
#' emoji font, and the rest. Arguments are recycled to the length of the `string`
#' vector.
#' 
#' @param string A character vector of strings that should be splitted.
#' @inheritParams match_font
#' @param path,index path an index of a font file to circumvent lookup based on 
#' family and style
#' 
#' @return A data.frame containing the following columns:
#' \describe{
#'   \item{string}{The substring containing a consecutive run of glyphs}
#'   \item{id}{The index into the original `string` vector that the substring is part of}
#'   \item{emoji}{A logical vector giving if the substring is a run of emojis or not}
#' }
#' 
#' @export
#' 
#' @examples 
#' emoji_string <- "This is a joke\U0001f642. It should be obvious from the smiley"
#' str_split_emoji(emoji_string)
#' 
str_split_emoji <- function(string, family = '', italic = FALSE, bold = FALSE, 
                            path = NULL, index = 0) {
  n_strings <- length(string)
  if (is.null(path)) {
    if (all(c(length(family), length(italic), length(bold)) == 1)) {
      loc <- match_font(family, italic, bold)
      path <- loc$path
      index <- loc$index
    } else {
      family <- rep_len(family, n_strings)
      italic <- rep_len(italic, n_strings)
      bold <- rep_len(bold, n_strings)
      loc <- Map(match_font, family = family, italic = italic, bold = bold)
      path <- vapply(loc, `[[`, character(1L), 1, USE.NAMES = FALSE)
      index <- vapply(loc, `[[`, integer(1L), 2, USE.NAMES = FALSE)
    }
  } else {
    if (!all(c(length(path), length(index)) == 1)) {
      path <- rep_len(path, n_strings)
      index <- rep_len(index, n_strings)
    }
  }
  if (!all(file.exists(path))) stop("path must point to a valid file", call. = FALSE)
  emoji_splitted <- emoji_split_c(as.character(string), path, index)
  groups <- diff(c(0, which(diff(emoji_splitted[[3]]) != 0), length(emoji_splitted[[3]])))
  groups <- rep(seq_along(groups), groups)
  groups <- paste0(emoji_splitted[[2]], '_', groups)
  string <- vapply(split(emoji_splitted[[1]], groups), intToUtf8, character(1))
  id <- vapply(split(emoji_splitted[[2]], groups), `[[`, integer(1), 1)
  emoji <- vapply(split(emoji_splitted[[3]], groups), `[[`, logical(1), 1)
  res <- data.frame(string = string, id = id + 1, emoji = emoji, stringsAsFactors = FALSE, row.names = NULL)
  class(res) <- c("tbl_df", "tbl", "data.frame")
  res
}

load_emoji_codes <- function() {
  load_emoji_codes_c(as.integer(all_emoji), as.integer(text_pres_emoji), 
    as.integer(base_mod_emoji))
}

#' Search font repositories for a font based on family name
#'
#' While it is often advisable to visit the webpage for a font repository when
#' looking for a font, in order to see examples etc, `search_web_fonts()`
#' provide a quick lookup based on family name in the repositories supported by
#' systemfonts (currently [Google Fonts](https://fonts.google.com) and
#' [Font Squirrel](https://fontsquirrel.com) - [Bunny Fonts](https://fonts.bunny.net/)
#' provide the same fonts as Google Fonts but doesn't have a search API). The
#' lookup is based on fuzzy matching provided by [utils::adist()] and the
#' matching parameters can be controlled through `...`
#'
#' @param family The font family name to look for
#' @param n_max The maximum number of matches to return
#' @inheritDotParams utils::adist -x -y
#'
#' @return A data.frame with the columns `family`, giving the family name of the
#' matched font, and `repository` giving the repository it was found in.
#'
#' @export
#'
#' @examples
#' # Requires an internet connection
#'
#' # search_web_fonts("Spectral")
#'
search_web_fonts <- function(family, n_max = 10, ...) {
  gf <- unique(get_google_fonts_registry()$family)
  fs <- unique(get_font_squirrel_registry()$family)
  all <- data.frame(
    family = c(gf, fs),
    repository = c(rep(
      c("Google Fonts/Bunny Fonts", "Font Squirrel"),
      c(length(gf), length(fs))
    ))
  )
  all[
    order(utils::adist(
      x = tolower(family),
      y = tolower(all$family),
      ...
    ))[seq_len(min(n_max, nrow(all)))],
  ]
}

#' Download and add web font
#'
#' In order to use a font in R it must first be made available locally. These
#' functions facilitate the download and registration of fonts from online
#' repositories.
#'
#' @param family The font family to download (case insensitive)
#' @param dir Where to download the font to. The default places it in your user
#' local font folder so that the font will be available automatically in new R
#' sessions. Set to `tempdir()` to only keep the font for the session.
#' @param woff2 Should the font be downloaded in the woff2 format (smaller and
#' more optimized)? Defaults to FALSE as the format is not supported on all
#' systems
#'
#' @return A logical invisibly indicating whether a font was found and
#' downloaded or not
#'
#' @name web-fonts
#' @rdname web-fonts
#'
NULL

#' @rdname web-fonts
#' @export
#'
get_from_google_fonts <- function(family, dir = "~/fonts", woff2 = FALSE) {
  fonts <- get_google_fonts_registry(woff2)
  match <- which(tolower(fonts$family) == tolower(family))

  if (length(match) == 0) {
    return(invisible(FALSE))
  }

  if (!dir.exists(dir)) dir.create(dir, recursive = TRUE)

  files <- fonts$url[match]
  download_name <- file.path(dir, fonts$file[match])
  success <- try(
    {
      if (capabilities("libcurl")) {
        utils::download.file(
          files,
          download_name,
          method = "libcurl",
          quiet = TRUE,
          mode = "wb"
        )
      } else {
        mapply(
          utils::download.file,
          url = files,
          destfile = download_name,
          quiet = TRUE,
          mode = "wb"
        )
      }
    },
    silent = TRUE
  )

  if (inherits(success, "try-error")) {
    return(invisible(FALSE))
  }

  add_fonts(download_name)
  invisible(TRUE)
}

#' @rdname web-fonts
#' @export
#'
get_from_font_squirrel <- function(family, dir = "~/fonts") {
  fonts <- get_font_squirrel_registry()
  match <- which(tolower(fonts$family) == tolower(family))

  if (length(match) == 0) {
    return(invisible(FALSE))
  }

  if (!dir.exists(dir)) dir.create(dir, recursive = TRUE)

  files <- fonts$url[match]
  download_name <- file.path(
    tempdir(check = TRUE),
    paste0(basename(fonts$url[match]), ".zip")
  )
  success <- try(
    {
      if (capabilities("libcurl")) {
        utils::download.file(
          files,
          download_name,
          method = "libcurl",
          quiet = TRUE
        )
      } else {
        mapply(
          utils::download.file,
          url = files,
          destfile = download_name,
          quiet = TRUE
        )
      }
    },
    silent = TRUE
  )

  if (inherits(success, "try-error")) {
    return(invisible(FALSE))
  }

  new_fonts <- unlist(mapply(
    utils::unzip,
    zipfile = download_name,
    MoreArgs = list(exdir = dir)
  ))
  is_font <- grepl("\\.(?:ttf|ttc|otf|otc|woff|woff2)$", tolower(new_fonts))
  unlink(new_fonts[!is_font])
  add_fonts(new_fonts[is_font])

  return(invisible(TRUE))
}
get_from_font_library <- function(family, dir = "~/fonts") {
  url <- import_from_font_library(family)
  if (length(url) == 0) {
    return(invisible(FALSE))
  }
  url <- readLines(url)
  urls <- grep("url\\(.*?\\)", url, value = TRUE)
  urls <- sub(".*url\\((.*?)\\).*", "\\1", urls)
  urls <- paste0("https://fontlibrary.org", gsub("'|\"", "", urls))

  if (!dir.exists(dir)) dir.create(dir, recursive = TRUE)

  download_name <- file.path(dir, basename(urls))
  success <- try(
    {
      if (capabilities("libcurl")) {
        utils::download.file(
          urls,
          download_name,
          method = "libcurl",
          quiet = TRUE
        )
      } else {
        mapply(
          utils::download.file,
          url = urls,
          destfile = download_name,
          quiet = TRUE
        )
      }
    },
    silent = TRUE
  )

  if (inherits(success, "try-error")) {
    return(invisible(FALSE))
  }

  add_fonts(download_name)

  return(invisible(TRUE))
}

#' Ensure font availability in a script
#'
#' When running a script on a different machine you are not always in control of
#' which fonts are installed on the system and thus how graphics created by the
#' script ends up looking. `require_font()` is a way to specify your font
#' requirements for a script. It will look at the available fonts and if the
#' required font family is not present it will attempt to fetch it from one of
#' the given repositories (in the order given). If that fails, it will either
#' throw an error or, if `fallback` is given, provide an alias for the fallback
#' so it maps to the required font.
#'
#' @param family The font family to require
#' @param fallback An available font to fall back to if `family` cannot be found
#' or downloaded
#' @param dir The location to put the font file downloaded from repositories
#' @param repositories The repositories to search for the font in case it is not
#' available on the system. They will be tried in the order given. Currently
#' `"Google Fonts"`, `"Font Squirrel"`, and `"Font Library"` is available.
#' @param error Should the function throw an error if unsuccessful?
#' @param verbose Should status messages be emitted?
#'
#' @return Invisibly `TRUE` if the font is available or `FALSE` if not (this can
#' only be returned if `error = FALSE`)
#'
#' @export
#'
#' @examples
#' # Should always work
#' require_font("sans")
#'
require_font <- function(
  family,
  fallback = NULL,
  dir = tempdir(),
  repositories = c("Google Fonts", "Font Squirrel", "Font Library"),
  error = TRUE,
  verbose = TRUE
) {
  if (tolower(family) %in% c("sans", "serif", "mono", "symbol"))
    return(invisible(TRUE))
  if (!is.character(family) || length(family) != 1) {
    stop("`family` must be a string")
  }
  if (
    !is.null(fallback) && (!is.character(fallback) || length(fallback) != 1)
  ) {
    stop("`family` must be a string")
  }
  fonts <- system_fonts()
  available <- which(tolower(fonts$family) == tolower(family))

  if (length(available) != 0) {
    if (verbose) {
      message(
        "`",
        family,
        "` available at ",
        paste0(unique(dirname(fonts$path[available])), collapse = ", ")
      )
    }
    return(invisible(TRUE))
  }

  success <- FALSE

  has_internet <- !inherits(
    suppressWarnings(try(
      readLines("https://8.8.8.8", n = 1L),
      silent = TRUE
    )),
    "try-error"
  )

  if (!has_internet) {
    if (verbose) {
      message("No internet connection. Can't search online repositories")
    }
  } else {
    for (repo in repositories) {
      if (verbose) message("Trying ", repo, "...", appendLF = FALSE)
      success <- switch(
        tolower(repo),
        "google fonts" = get_from_google_fonts(family, dir),
        "font squirrel" = get_from_font_squirrel(family, dir),
        "font library" = get_from_font_library(family, dir),
        FALSE
      )
      if (verbose) {
        if (success) {
          message(" Found! Downloading font to ", dir)
        } else {
          message("Not found.")
        }
      }
      if (success) break
    }
  }

  if (!success) {
    if (is.null(fallback)) {
      if (error)
        stop(paste0(
          "Required font: ",
          family,
          ", is not available on the system"
        ))
    } else {
      message(
        "Required font: `",
        family,
        "`, is not available on the system. Adding alias to `",
        fallback,
        "`"
      )
      register_variant(family, fallback)
      success <- TRUE
    }
  }

  invisible(success)
}

#' Create import specifications for web content
#'
#' If you create content in a text-based format such as HTML or SVG you need to
#' make sure that the font is available on the computer where it is viewed. This
#' can be achieved through the use of stylesheets that can either be added with
#' a `<link>` tag or inserted with an `@import` statement. This function
#' facilitates the creation of either of these (or the bare URL to the
#' stylesheet). It can rely on the Bunny Fonts, Google Fonts and/or Font Library
#' repositories for serving the fonts. If the requested font is not found it can
#' optionally hard code the data into the stylesheet.
#'
#' @inheritParams match_fonts
#' @param ... Additional arguments passed on to the specific functions for the
#' repositories. Currently:
#' * **Google Fonts and Bunny Fonts:**
#'   - `text` A piece of text containing the glyphs required. Using this can
#'     severely cut down on the size of the required download
#'   - `display` One of `"auto"`, `"block"`, `"swap"`, `"fallback"`, or
#'     `"optional"`. Controls how the text is displayed while the font is
#'     downloading.
#' @param type The type of return value. `"url"` returns the bare url pointing
#' to the style sheet. `"import"` returns the stylesheet as an import statement
#' (`@import url(<url>)`). `"link"` returns the stylesheet as a link tag
#' (`<link rel="stylesheet" href="<url>"/>`)
#' @param may_embed Logical. Should fonts that can't be found in the provided
#' repositories be embedded as data-URLs. This is only possible if the font is
#' available locally and in a `woff2`, `woff`, `otf`, or `ttf` file.
#' @param repositories The repositories to try looking for the font. Currently
#' `"Bunny Fonts"`, `"Google Fonts"`, and `"Font Library"` are supported. Set
#' this to `NULL` together with `may_embed = TRUE` to force embedding of the
#' font data.
#'
#' @return A character vector with stylesheet specifications according to `type`
#' @export
#'
fonts_as_import <- function(
  family,
  italic = NULL,
  weight = NULL,
  width = NULL,
  ...,
  type = c("url", "import", "link"),
  may_embed = TRUE,
  repositories = c("Bunny Fonts", "Google Fonts", "Font Library")
) {
  import <- character(0)
  type <- match.arg(type)
  if (may_embed) repositories <- c(repositories, "local")

  all_families <- family

  for (repo in repositories) {
    fonts <- switch(
      tolower(repo),
      "bunny fonts" = import_from_bunny_fonts(
        family,
        italic = italic,
        weight = weight,
        width = width,
        ...
      ),
      "google fonts" = import_from_google_fonts(
        family,
        italic = italic,
        weight = weight,
        width = width,
        ...
      ),
      "font library" = import_from_font_library(family, ...),
      "local" = import_embedded(
        family,
        italic = italic,
        weight = weight,
        width = width,
        ...
      ),
      NULL
    )
    if (!is.null(fonts)) {
      import <- c(import, fonts)
      missing <- attr(fonts, "no_match")

      family <- family[missing]
      if (!is.null(italic)) italic <- italic[missing]
      if (!is.null(width)) width <- width[missing]
      if (!is.null(weight)) weight <- weight[missing]
    }
    if (length(family) == 0) break
  }

  if (length(family) != 0) {
    warning("No import found for ", paste(family, collapse = ", "))
  }

  imported <- unique(setdiff(all_families, family))
  for (f in imported) {
    success <- require_font(imported, error = FALSE, verbose = FALSE)
    if (!success) {
      warning(
        "An import URL for ",
        f,
        " could be made but the font could not be made avialable on the system"
      )
    }
  }

  switch(
    type,
    import = paste0("@import url('", import, "');"),
    link = paste0('<link rel="stylesheet" href="', import, '"/>'),
    url = import
  )
}

import_from_google_fonts <- function(
  family,
  italic = NULL,
  weight = NULL,
  width = NULL,
  ...,
  text = NULL,
  display = "swap"
) {
  n_fonts <- max(length(family), length(italic), length(weight), length(width))
  family <- rep_len(family, n_fonts)
  if (!is.null(italic)) italic <- rep_len_default(as.integer(italic), n_fonts, 0L)
  if (!is.null(weight))
    weight <- rep_len_default(systemfonts::as_font_weight(weight), n_fonts, 400)
  if (!is.null(width))
    width <- rep_len_default(systemfonts::as_font_width(width), n_fonts, 0)

  clusters <- split(seq_along(family), family)

  possible_fonts <- unique(get_google_fonts_registry()$family)

  fonts <- lapply(clusters, function(i) {
    fam <- family[i[1]]
    fam <- match(tolower(fam), tolower(possible_fonts))
    if (is.na(fam)) return()
    fam <- possible_fonts[fam]
    fam <- paste0("family=", gsub(" ", "+", family[i[1]]))

    spec <- list()
    spec$ital <- if (!is.null(italic)) as.character(unique(range(italic[i])))
    if (isTRUE(spec$ital == 0L)) spec$ital <- NULL
    spec$wdth <- if (!is.null(width))
      paste0(unique(range(width[i])), collapse = "..")
    spec$wght <- if (!is.null(weight))
      paste0(unique(range(weight[i])), collapse = "..")
    spec <- spec[lengths(spec) != 0]
    if (length(spec) != 0) {
      n_specs <- max(lengths(spec))
      spec <- lapply(spec, rep_len, n_specs)
      val <- paste0(
        vapply(
          seq_len(n_specs),
          function(i) {
            paste0(vapply(spec, `[[`, character(1), i), collapse = ",")
          },
          character(1)
        ),
        collapse = ";"
      )
      spec <- paste0(names(spec), collapse = ",")
      fam <- paste0(fam, ":", spec, "@", val)
    }
    fam
  })
  missing <- unlist(clusters[lengths(fonts) == 0]) %||% integer()
  fonts <- paste0(unlist(fonts), collapse = "&")

  display <- match.arg(
    display,
    c("auto", "block", "swap", "fallback", "optional")
  )

  url <- paste0("https://fonts.googleapis.com/css2?", fonts)

  if (display != "auto") {
    url <- paste0(url, "&display=", display)
  }

  if (!is.null(text)) {
    url <- paste0(url, "&text=", utils::URLencode(text))
  }

  success <- try(suppressWarnings(readLines(url, n = 1)), silent = TRUE)
  if (inherits(success, "try-error")) {
    url <- character(0)
    missing <- seq_along(family)
  }
  structure(url %||% character(), no_match = missing)
}
import_from_bunny_fonts <- function(
  family,
  italic = NULL,
  weight = NULL,
  width = NULL,
  ...,
  text = NULL,
  display = "swap"
) {
  url <- import_from_google_fonts(
    family = family,
    italic = italic,
    weight = weight,
    width = width,
    ...,
    text = text,
    display = display
  )
  success <- try(suppressWarnings(readLines(url, n = 1)), silent = TRUE)
  if (
    inherits(success, "try-error") || any(grepl("Error: API Error", success))
  ) {
    return(structure(character(), no_match = seq_along(family)))
  }
  structure(
    sub(
      "https://fonts.googleapis.com/css2?",
      "https://fonts.bunny.net/css2?",
      url,
      fixed = TRUE
    ),
    no_match = attr(url, "no_match")
  )
}
import_from_font_library <- function(family, ...) {
  family <- gsub(" ", "-", tolower(family))
  u_fam <- unique(family)
  names(u_fam) <- u_fam
  fonts <- lapply(u_fam, function(name) {
    url <- paste0("https://fontlibrary.org/face/", name)
    if (length(readLines(url, n = 1)) != 0) {
      url
    } else {
      NULL
    }
  })
  missing <- which(family %in% names(fonts[lengths(fonts) == 0]))
  fonts <- unlist(fonts)

  structure(fonts %||% character(), no_match = missing, names = NULL)
}
import_embedded <- function(
  family,
  italic = NULL,
  weight = NULL,
  width = NULL,
  ...
) {
  fonts <- match_fonts(
    family,
    italic = italic %||% FALSE,
    weight = weight %||% "normal",
    width = width %||% "undefined"
  )
  fonts <- lapply(seq_along(family), function(i) {
    found <- font_info(path = fonts$path[i], index = fonts$index[i])
    if (tolower(family[i]) != tolower(found$family)) return()
    ext <- tools::file_ext(fonts$path[i])
    if (!ext %in% c("woff2", "woff", "otf", "ttf")) {
      warning(ext, "-files cannot be embedded (", family[i], ")")
      return()
    }
    format <- switch(
      ext,
      woff2 = 'format("woff2")',
      woff = 'format("woff")',
      otf = 'format("opentype")',
      ttf = 'format("truetype")'
    )
    src <- paste0(
      "url(data:font/",
      ext,
      ";charset=utf-8;base64,",
      base64enc::base64encode(fonts$path[i]),
      ") ",
      format
    )
    features <- paste0(
      '"',
      fonts$features[[i]][[1]],
      '" ',
      fonts$features[[i]][[2]],
      collapse = ", "
    )
    # fmt: skip
    x <- paste0(
      '@font-face {\n',
      '  font-family: "', found$family, '";\n',
      '  src: ', src, ';\n',
      if (length(fonts$features[[i]]) != 0) paste0(
      '  font-feature-settings: ', features, ';\n'),
      if (!is.na(found$width)) paste0(
      '  font-stretch: ', sub("(ra|mi)", "\\1-", found$width), ';\n'),
      if (!is.na(found$weight)) paste0(
      '  font-weight: ', as_font_weight(found$weight), ';\n'),
      '  font-style: ', if (found$italic) "italic" else "normal", ';\n',
      '}'
    )
    paste0("data:text/css,", utils::URLencode(x))
  })
  missing <- which(lengths(fonts) == 0)
  fonts <- unlist(fonts)

  structure(fonts %||% character(), no_match = missing)
}

# REGISTRIES -------------------------------------------------------------------

google_fonts_registry <- new.env(parent = emptyenv())
get_google_fonts_registry <- function(woff2 = FALSE) {
  loc <- if (woff2) "woff2" else "ttf"
  if (is.null(google_fonts_registry[[loc]])) {
    url <- "https://www.googleapis.com/webfonts/v1/webfonts?key=AIzaSyBkOYsZREsyZWvbSR_d03SI5XX30cIapYo&sort=popularity"
    if (woff2) {
      url <- paste0(url, "&capability=WOFF2")
    }
    fonts <- jsonlite::read_json(url)$items
    google_fonts_registry[[loc]] <- do.call(
      rbind,
      lapply(fonts, function(x) {
        x <- data.frame(
          family = x$family,
          variant = unlist(x$variants),
          url = unlist(x$files),
          file = NA_character_,
          version = x$version,
          modified = x$lastModified,
          category = I(rep(list(x$category), length(x$variants)))
        )
        x$file <- paste0(
          x$family,
          "-",
          x$variant,
          sub("^.*(\\.\\w+)$", "\\1", x$url)
        )
        attr(x, "row.names") <- .set_row_names(nrow(x))
        x
      })
    )
  }
  google_fonts_registry[[loc]]
}

font_squirrel_registry <- new.env(parent = emptyenv())
get_font_squirrel_registry <- function() {
  if (is.null(font_squirrel_registry$fonts)) {
    fonts <- jsonlite::read_json(
      "https://www.fontsquirrel.com/api/fontlist/all"
    )
    font_squirrel_registry$fonts <- do.call(
      rbind,
      lapply(fonts, function(x) {
        x <- data.frame(
          family = x$family_name,
          n_variants = x$family_count,
          url = paste0(
            "https://www.fontsquirrel.com/fonts/download/",
            x$family_urlname
          ),
          category = I(list(x$classification))
        )
        attr(x, "row.names") <- .set_row_names(nrow(x))
        x
      })
    )
  }
  font_squirrel_registry$fonts
}

release_bullets <- function() {
  c(
    '`rhub::check_on_solaris(env_vars = c("_R_CHECK_FORCE_SUGGESTS_" = "false"))`',
    '`rhub::check_with_valgrind(env_vars = c(VALGRIND_OPTS = "--leak-check=full --track-origins=yes"))`'
  )
}

.onLoad <- function(...) {
  load_emoji_codes()
  scan_local_fonts()
  windows_workaround()
}

warn_env <- new.env(parent = emptyenv())
warn_env$warned <- FALSE

#' Get location of the fallback font
#'
#' @export
#' @keywords internal
get_fallback <- function() {
  if (!warn_env$warned) {
    warning(
      "No fonts detected on your system. Using an empty font.",
      call. = FALSE
    )
    warn_env$warned <- TRUE
  }
  list(
    system.file("unfont.ttf", package = "systemfonts"),
    0L
  )
}

# See https://github.com/r-lib/textshaping/issues/36
windows_workaround <- function(){
  # This dll needs to be loaded before the textshaping pkg so we do it here.
  # It is harmless otherwise (it is also loaded by e.g. RGui in Windows)
  if(.Platform$OS.type == 'windows'){
    if(file.exists("C:\\Windows\\System32\\TextShaping.dll")){
      dyn.load("C:\\Windows\\System32\\TextShaping.dll")
    }    
  }
}

`%||%` <- function(a, b) if (is.null(a)) b else a

rep_len_default <- function(x, length.out, default) {
  if (length(x) == 0) x <- default
  rep_len(x, length.out)
}

#' Define OpenType font feature settings
#' 
#' This function encapsulates the specification of OpenType font features. Some
#' specific features have named arguments, but all available features can be
#' set by using its specific 4-letter tag For a list of the 4-letter tags
#' available see e.g. the overview on 
#' [Wikipedia](https://en.wikipedia.org/wiki/List_of_typographic_features).
#' 
#' @param ligatures Settings related to ligatures. One or more types of 
#' ligatures to turn on (see details).
#' @param letters Settings related to the appearance of single 
#' letters (as opposed to ligatures that substitutes multiple letters). See 
#' details for supported values.
#' @param numbers Settings related to the appearance of numbers. See details for
#' supported values. 
#' @param ... key-value pairs with the key being the 4-letter tag and the value
#' being the setting (usually `TRUE` to turn it on).
#' 
#' @details 
#' OpenType features are defined by a 4-letter tag along with an integer value.
#' Often that value is a simple `0` (off) or `1` (on), but some features support
#' additional values, e.g. stylistic alternates (`salt`) where a font may 
#' provide multiple variants of a letter and the value will be used to chose 
#' which one to use.
#' 
#' Common features related to appearance may be given with a long form name to
#' either the `ligatures`, `letters`, or `numbers` argument to avoid remembering
#' the often arbitrary 4-letter tag. Providing a long form name is the same as
#' setting the tag to `1` and can thus not be used to set tags to other values.
#' 
#' The possible long form names are given below with the tag in parenthesis:
#' 
#' **Ligatures**  
#' - `standard` (*liga*): Turns on standard multiple letter substitution
#' - `historical` (*hlig*): Use obsolete historical ligatures
#' - `contextual` (*clig*): Apply secondary ligatures based on the character 
#'   patterns surrounding the potential ligature
#' - `discretionary` (*dlig*): Use ornamental ligatures
#' 
#' **Letters**  
#' - `swash` (*cswh*): Use contextual swashes (ornamental decorations)
#' - `alternates` (*calt*): Use alternate letter forms based on the sourrounding
#'   pattern
#' - `historical` (*hist*): Use obsolete historical forms of the letters
#' - `localized` (*locl*): Use alternate forms preferred by the script language 
#' - `randomize` (*rand*): Use random variants of the letters (e.g. to mimick
#'   handwriting)
#' - `alt_annotation` (*nalt*): Use alternate annotations (e.g. circled digits)
#' - `stylistic` (*salt*): Use a stylistic alternative form of the letter
#' - `subscript` (*subs*): Set letter in subscript
#' - `superscript` (*sups*): Set letter in superscript
#' - `titling` (*titl*): Use letter forms well suited for large text and titles
#' - `small_caps` (*smcp*): Use small caps variants of the letters
#' 
#' **Numbers**
#' - `lining` (*lnum*): Use number variants that rest on the baseline
#' - `oldstyle` (*onum*): Use old style numbers that use descender and ascender
#'   for various numbers
#' - `proportional` (*pnum*): Let numbers take up width based on the visual 
#'   width of the glyph
#' - `tabular` (*tnum*): Enforce all numbers to take up the same width
#' - `fractions` (*frac*): Convert numbers separated by `/` into a fraction 
#'   glyph
#' - `fractions_alt` (*afrc*): Use alternate fraction form with a horizontal 
#'   divider
#' 
#' @return A `font_feature` object
#' 
#' @export
#' 
#' @examples 
#' font_feature(letters = "stylistic", numbers = c("lining", "tabular"))
#' 
#' # Use the tag directly to access additional stylistic variants
#' font_feature(numbers = c("lining", "tabular"), salt = 2)
#' 
font_feature <- function(ligatures = NULL, letters = NULL, numbers = NULL , ...) {
  features <- list(...)
  if (!is.null(ligatures)) {
    for (lig in ligatures) {
      features[get_ligature_tag(lig)] <- 1L
    }
  }
  if (!is.null(letters)) {
    for (let in letters) {
      features[get_letter_tag(let)] <- 1L
    }
  }
  if (!is.null(numbers)) {
    for (num in numbers) {
      features[get_number_tag(num)] <- 1L
    }
  }
  if (length(features) == 0) {
    features <- list(character(), integer())
  } else {
    feature_names <- names(features)
    if (any(vapply(features, length, integer(1)) != 1)) {
      stop("A feature setting can only be of length 1", call. = FALSE)
    }
    if (anyDuplicated(feature_names)) {
      stop("Features can only be given once", call. = FALSE)
    }
    if (any(nchar(feature_names) != 4)) {
      stop("Feature tags must be 4 character long", call. = FALSE)
    }
    feature_settings <- vapply(features, as.integer, integer(1))
    features <- list(feature_names, feature_settings)
  }
  class(features) <- "font_feature"
  features
}
is_font_feature <- function(x) inherits(x, "font_feature")

length.font_feature <- function(x) {
  length(x[[1]])
}
#' @export
print.font_feature <- function(x, ...) {
  if (length(x) == 0) {
    cat("An empty font_feature object\n")
  } else {
    cat("A list of OpenType font feature settings\n")
    cat(paste(paste0("- ", x[[1]], ": ", x[[2]]), collapse = "\n"))
  }
  invisible(x)
}

get_ligature_tag <- function(x) {
  name <- c("standard", "historical", "contextual", "discretionary")
  tags <- c("liga",     "hlig",       "clig",       "dlig"         )
  ind <- match(tolower(x), name)
  if (is.na(ind)) {
    stop("No ligature setting called '", x, "'. Use one of ", paste(name, collapse = ", "), call. = FALSE)
  }
  tags[ind]
}
get_letter_tag <- function(x) {
  name <- c("swash", "alternates", "historical", "localized", "randomize", "alt_annotation", "stylistic", "subscript", "superscript", "titling", "small_caps")
  tags <- c("cswh",  "calt",       "hist",       "locl",      "rand",      "nalt",           "salt",      "subs",      "sups",        "titl",    "smcp"      )
  ind <- match(tolower(x), name)
  if (is.na(ind)) {
    stop("No letter setting called '", x, "'. Use one of ", paste(name, collapse = ", "), call. = FALSE)
  }
  tags[ind]
}
get_number_tag <- function(x) {
  name <- c("lining", "oldstyle", "proportional", "tabular", "fractions", "fractions_alt")
  tags <- c("lnum",   "onum",     "pnum",         "tnum",    "frac",      "afrc"         )
  ind <- match(tolower(x), name)
  if (is.na(ind)) {
    stop("No number setting called '", x, "'. Use one of ", paste(name, collapse = ", "), call. = FALSE)
  }
  tags[ind]
}

#' Calculate glyph positions for strings
#'
#' Do basic text shaping of strings. This function will use freetype to
#' calculate advances, doing kerning if possible. It will not perform any font
#' substitution or ligature resolving and will thus be much in line with how
#' the standard graphic devices does text shaping. Inputs are recycled to the
#' length of `strings`.
#'
#' @param strings A character vector of strings to shape
#' @param id A vector grouping the strings together. If strings share an id the
#' shaping will continue between strings
#' @inheritParams font_info
#' @param lineheight A multiplier for the lineheight
#' @param align Within text box alignment, either `'left'`, `'center'`, or
#' `'right'`
#' @param hjust,vjust The justification of the textbox surrounding the text
#' @param max_width The requested with of the string in inches. Setting this to
#' something other than `NA` will turn on word wrapping.
#' @param tracking Tracking of the glyphs (space adjustment) measured in 1/1000
#' em.
#' @param indent The indent of the first line in a paragraph measured in inches.
#' @param hanging The indent of the remaining lines in a paragraph measured in
#' inches.
#' @param space_before,space_after The spacing above and below a paragraph,
#' measured in points
#' @param path,index path an index of a font file to circumvent lookup based on
#' family and style
#'
#' @return
#' A list with two element: `shape` contains the position of each glyph,
#' relative to the origin in the enclosing textbox. `metrics` contain metrics
#' about the full strings.
#'
#' `shape` is a data.frame with the following columns:
#' \describe{
#'   \item{glyph}{The glyph as a character}
#'   \item{index}{The index of the glyph in the font file}
#'   \item{metric_id}{The index of the string the glyph is part of (referencing a row in the `metrics` data.frame)}
#'   \item{string_id}{The index of the string the glyph came from (referencing an element in the `strings` input)}
#'   \item{x_offset}{The x offset in pixels from the origin of the textbox}
#'   \item{y_offset}{The y offset in pixels from the origin of the textbox}
#'   \item{x_mid}{The x offset in pixels to the middle of the glyph, measured from the origin of the glyph}
#' }
#'
#' `metrics` is a data.frame with the following columns:
#' \describe{
#'   \item{string}{The text the string consist of}
#'   \item{width}{The width of the string}
#'   \item{height}{The height of the string}
#'   \item{left_bearing}{The distance from the left edge of the textbox and the leftmost glyph}
#'   \item{right_bearing}{The distance from the right edge of the textbox and the rightmost glyph}
#'   \item{top_bearing}{The distance from the top edge of the textbox and the topmost glyph}
#'   \item{bottom_bearing}{The distance from the bottom edge of the textbox and the bottommost glyph}
#'   \item{left_border}{The position of the leftmost edge of the textbox related to the origin}
#'   \item{top_border}{The position of the topmost edge of the textbox related to the origin}
#'   \item{pen_x}{The horizontal position of the next glyph after the string}
#'   \item{pen_y}{The vertical position of the next glyph after the string}
#' }
#'
#' @export
#'
#' @examples
#' string <- "This is a long string\nLook; It spans multiple lines\nand all"
#'
#' # Shape with default settings
#' shape_string(string)
#'
#' # Mix styles within the same string
#' string <- c(
#'   "This string will have\na ",
#'   "very large",
#'   " text style\nin the middle"
#' )
#'
#' shape_string(string, id = c(1, 1, 1), size = c(12, 24, 12))
#'
shape_string <- function(
  strings,
  id = NULL,
  family = '',
  italic = FALSE,
  weight = "normal",
  width = "undefined",
  size = 12,
  res = 72,
  lineheight = 1,
  align = 'left',
  hjust = 0,
  vjust = 0,
  max_width = NA,
  tracking = 0,
  indent = 0,
  hanging = 0,
  space_before = 0,
  space_after = 0,
  path = NULL,
  index = 0,
  bold = deprecated()
) {
  n_strings = length(strings)
  if (is.null(id)) id <- seq_len(n_strings)
  id <- rep_len_default(id, n_strings, seq_len(n_strings))
  id <- match(id, unique(id))
  if (anyNA(id)) {
    stop('id must be a vector of valid integers', call. = FALSE)
  }
  ido <- order(id)
  id <- id[ido]
  strings <- as.character(strings)[ido]

  if (is.null(path)) {
    if (lifecycle::is_present(bold)) {
      lifecycle::deprecate_soft(
        "1.2.4",
        "shape_string(bold)",
        "shape_string(weight)"
      )
      weight <- ifelse(bold, "bold", "normal")
    }
    fonts <- match_fonts(
      family = rep_len_default(family, n_strings, ''),
      italic = rep_len_default(italic, n_strings, FALSE),
      weight = rep_len_default(weight, n_strings, "normal"),
      width = rep_len_default(width, n_strings, "undefined")
    )
    path <- fonts$path[ido]
    index <- fonts$index[ido]
  } else {
    if (!all(c(length(path), length(index)) == 1)) {
      path <- rep_len(path, n_strings)[ido]
      index <- rep_len(index, n_strings)[ido]
    }
  }
  if (length(size) != 1) size <- rep_len_default(size, n_strings, 12)[ido]
  if (length(res) != 1) res <- rep_len_default(res, n_strings, 72)[ido]
  if (length(lineheight) != 1) lineheight <- rep_len_default(lineheight, n_strings, 1)[ido]
  align <- match.arg(align, c('left', 'center', 'right'), TRUE)
  align <- match(align, c('left', 'center', 'right'))
  if (length(align) != 1) align <- rep_len_default(align, n_strings, 1)[ido]
  if (length(hjust) != 1) hjust <- rep_len_default(hjust, n_strings, 0)[ido]
  if (length(vjust) != 1) vjust <- rep_len_default(vjust, n_strings, 0)[ido]
  if (length(max_width) != 1) max_width <- rep_len_default(max_width, n_strings, NA)[ido]
  max_width[is.na(max_width)] <- -1
  if (length(tracking) != 1) tracking <- rep_len_default(tracking, n_strings, 0)[ido]
  if (length(indent) != 1) indent <- rep_len_default(indent, n_strings, 0)[ido]
  if (length(hanging) != 1) hanging <- rep_len_default(hanging, n_strings, 0)[ido]
  if (length(space_before) != 1) {
    space_before <- rep_len_default(space_before, n_strings, 0)[ido]
  }
  if (length(space_after) != 1) {
    space_after <- rep_len_default(space_after, n_strings, 0)[ido]
  }

  max_width <- max_width * res
  indent <- indent * res
  hanging <- hanging * res

  if (!all(file.exists(path))) {
    stop("path must point to a valid file", call. = FALSE)
  }
  shape <- get_string_shape_c(
    strings,
    id,
    path,
    as.integer(index),
    as.numeric(size),
    as.numeric(res),
    as.numeric(lineheight),
    as.integer(align) - 1L,
    as.numeric(hjust),
    as.numeric(vjust),
    as.numeric(max_width),
    as.numeric(tracking),
    as.numeric(indent),
    as.numeric(hanging),
    as.numeric(space_before),
    as.numeric(space_after)
  )

  shape$metrics$string <- vapply(
    split(strings, id),
    paste,
    character(1),
    collapse = ''
  )
  shape$shape$string_id <- ido[shape$shape$string_id]
  shape$shape <- shape$shape[order(shape$shape$string_id), , drop = FALSE]
  shape$shape$glyph <- intToUtf8(shape$shape$glyph, multiple = TRUE)
  shape$shape$x_offset <- shape$shape$x_offset * (72 / res)
  shape$shape$y_offset <- shape$shape$y_offset * (72 / res)
  shape$shape$x_midpoint <- shape$shape$x_midpoint * (72 / res)
  shape
}
#' Calculate the width of a string, ignoring new-lines
#'
#' This is a very simple alternative to [shape_string()] that simply calculates
#' the width of strings without taking any newline into account. As such it is
#' suitable to calculate the width of words or lines that has already been
#' splitted by `\n`. Input is recycled to the length of `strings`.
#'
#' @inheritParams font_info
#' @param strings A character vector of strings
#' @param include_bearing Logical, should left and right bearing be included in
#' the string width?
#'
#' @return A numeric vector giving the width of the strings in pixels. Use the
#' provided `res` value to convert it into absolute values.
#'
#' @export
#'
#' @examples
#' strings <- c('A short string', 'A very very looong string')
#' string_width(strings)
#'
string_width <- function(
  strings,
  family = '',
  italic = FALSE,
  weight = "normal",
  width = "undefined",
  size = 12,
  res = 72,
  include_bearing = TRUE,
  path = NULL,
  index = 0,
  bold = deprecated()
) {
  n_strings <- length(strings)
  if (is.null(path)) {
    if (lifecycle::is_present(bold)) {
      lifecycle::deprecate_soft(
        "1.2.4",
        "string_width(bold)",
        "string_width(weight)"
      )
      weight <- ifelse(bold, "bold", "normal")
    }
    fonts <- match_fonts(
      family = rep_len_default(family, n_strings, ''),
      italic = rep_len_default(italic, n_strings, FALSE),
      weight = rep_len_default(weight, n_strings, "normal"),
      width = rep_len_default(width, n_strings, "undefined")
    )
    path <- fonts$path
    index <- fonts$index
  } else {
    if (!all(c(length(path), length(index)) == 1)) {
      path <- rep_len(path, n_strings)
      index <- rep_len(index, n_strings)
    }
  }
  if (length(size) != 1) size <- rep_len_default(size, n_strings, 12)
  if (length(res) != 1) res <- rep_len_default(res, n_strings, 72)
  if (length(include_bearing) != 1) {
    include_bearing <- rep_len_default(include_bearing, n_strings, TRUE)
  }
  if (!all(file.exists(path))) {
    stop("path must point to a valid file", call. = FALSE)
  }
  get_line_width_c(
    as.character(strings),
    path,
    as.integer(index),
    as.numeric(size),
    as.numeric(res),
    as.logical(include_bearing)
  )
}

#' Find a system font by name and style
#'
#' This function locates the font file (and index) best matching a name and
#' optional style. A font file will be returned even if a perfect match
#' isn't found, but it is not necessarily similar to the requested family and
#' it should not be relied on for font substitution. The aliases `"sans"`,
#' `"serif"`, `"mono"`, `"symbol"`, and `"emoji"` match to their respective
#' system defaults (`""` is equivalent to `"sans"`). `match_font()` has been
#' deprecated in favour of `match_fonts()` which provides vectorisation, as well
#' as querying for different weights (rather than just "normal" and "bold") as
#' well as different widths.
#'
#' # Font matching
#' During font matching, systemfonts has to look in three different locations.
#' The font registry (populated by [register_font()]/[register_variant()]), the
#' local fonts (populated with [add_fonts()]/[scan_local_fonts()]), and the
#' fonts installed on the system. It does so in that order: registry > local >
#' installed.
#'
#' The matching performed at each step also differs. The fonts in the registry
#' is only matched by family name. The local fonts are matched based on all the
#' provided parameters (family, weight, italic, etc) in a way that is local to
#' systemfonts, but try to emulate the system native matching. The installed
#' fonts are matched using the system native matching functionality on macOS and
#' Linux. On Windows the installed fonts are read from the system registry and
#' matched using the same approach as for local fonts. Matching will always find
#' a font no matter what you throw at it, defaulting to "sans" if nothing else
#' is found.
#'
#' @param family The name of the font families to match
#' @param italic logical indicating the font slant
#' @param weight The weight to query for, either in numbers (`0`, `100`, `200`,
#' `300`, `400`, `500`, `600`, `700`, `800`, or `900`) or strings (`"undefined"`,
#' `"thin"`, `"ultralight"`, `"light"`, `"normal"`, `"medium"`, `"semibold"`,
#' `"bold"`, `"ultrabold"`, or `"heavy"`). `NA` will be interpreted as
#' `"undefined"`/`0`
#' @param width The width to query for either in numbers (`0`, `1`, `2`,
#' `3`, `4`, `5`, `6`, `7`, `8`, or `9`) or strings (`"undefined"`,
#' `"ultracondensed"`, `"extracondensed"`, `"condensed"`, `"semicondensed"`,
#' `"normal"`, `"semiexpanded"`, `"expanded"`, `"extraexpanded"`, or
#' `"ultraexpanded"`). `NA` will be interpreted as `"undefined"`/`0`
#' @param bold logical indicating whether the font weight
#'
#' @return A list containing the paths locating the font files, the 0-based
#' index of the font in the files and the features for the font in case a
#' registered font was located.
#'
#' @export
#'
#' @examples
#' # Get the system default sans-serif font in italic
#' match_fonts('sans', italic = TRUE)
#'
#' # Try to match it to a thin variant
#' match_fonts(c('sans', 'serif'), weight = "thin")
#'
match_fonts <- function(
  family,
  italic = FALSE,
  weight = "normal",
  width = "undefined"
) {
  if (!is.character(family) || anyNA(family))
    stop("`family` must be a character vector without NA values", call. = FALSE)
  weight <- as_font_weight(weight)
  width <- as_font_width(width)
  n_max <- max(length(family), length(italic), length(weight), length(width))
  locate_fonts_c(
    rep_len_default(family, n_max, ""),
    rep_len_default(as.numeric(italic), n_max, 0),
    rep_len_default(weight, n_max, 400),
    rep_len_default(width, n_max, 0)
  )
}
#' @rdname match_fonts
#' @export
match_font <- function(family, italic = FALSE, bold = FALSE) {
  lifecycle::deprecate_soft("1.1.0", "match_font()", "match_fonts()")
  if (!is.character(family)) stop("family must be a string", call. = FALSE)
  match_font_c(family, as.logical(italic), as.logical(bold))
}


weights <- c(
  "undefined",
  "thin",
  "ultralight",
  "light",
  "normal",
  "medium",
  "semibold",
  "bold",
  "ultrabold",
  "heavy"
)
#' Convert weight and width to numerics
#'
#' It is often more natural to describe font weight and width with names rather
#' than numbers (e.g. "bold" or "condensed"), but underneath these names are
#' matched to numeric values. These two functions are used to retrieve the
#' numeric counterparts to names
#'
#' @param weight,width character vectors with valid names for weight or width
#'
#' @return An integer vector matching the length of the input
#'
#' @keywords internal
#'
#' @export
#'
#' @examples
#' as_font_weight(
#'   c("undefined", "thin", "ultralight", "light", "normal", "medium", "semibold",
#'     "bold", "ultrabold", "heavy")
#' )
#'
as_font_weight <- function(weight) {
  if (is.factor(weight)) weight <- as.character(weight)
  if (is.logical(weight) && all(is.na(weight))) weight <- "undefined"
  if (is.character(weight)) {
    weight[is.na(weight)] <- "undefined"
    weight <- match(tolower(weight), weights)
    if (anyNA(weight)) {
      stop(
        paste0(
          "`weight` must be one of ",
          paste('"', weights[1:9], '"', collapse = ", ", sep = ''),
          ', or "',
          weights[10],
          '"'
        ),
        call. = FALSE
      )
    }
    weight <- (weight - 1) * 100
  } else if (is.numeric(weight) || is.integer(weight)) {
    weight[is.na(weight)] <- 0
  } else {
    stop("Can't convert this object to a font weight", call. = FALSE)
  }
  as.numeric(weight)
}
widths <- c(
  "undefined",
  "ultracondensed",
  "extracondensed",
  "condensed",
  "semicondensed",
  "normal",
  "semiexpanded",
  "expanded",
  "extraexpanded",
  "ultraexpanded"
)
#' @rdname as_font_weight
#' @export
#'
#' @examples
#' as_font_width(
#'   c("undefined", "ultracondensed", "extracondensed", "condensed", "semicondensed",
#'   "normal", "semiexpanded", "expanded", "extraexpanded", "ultraexpanded")
#' )
#'
as_font_width <- function(width) {
  if (is.factor(width)) width <- as.character(width)
  if (is.logical(width) && all(is.na(width))) width <- "undefined"
  if (is.character(width)) {
    width[is.na(width)] <- "undefined"
    width <- match(tolower(width), widths)
    if (anyNA(width)) {
      stop(
        paste0(
          "`width` must be one of ",
          paste('"', widths[1:9], '"', collapse = ", ", sep = ''),
          ', or "',
          widths[10],
          '"'
        ),
        call. = FALSE
      )
    }
    width <- width - 1
  } else if (is.numeric(width) || is.integer(width)) {
    width[is.na(width)] <- 0
  } else {
    stop("Can't convert this object to a font width", call. = FALSE)
  }
  as.numeric(width)
}

#' Get the fallback font for a given string
#'
#' A fallback font is a font to use as a substitute if the chosen font does not
#' contain the requested characters. Using font fallbacks means that the user
#' doesn't have to worry about mixing characters from different scripts or
#' mixing text and emojies. Fallback is calculated for the full string and the
#' result is platform specific. If no font covers all the characters in the
#' string an undefined "best match" is returned. The best approach is to figure
#' out which characters are not covered by your chosen font and figure out
#' fallbacks for these, rather than just request a fallback for the full string.
#'
#' @param string The strings to find fallbacks for
#' @inheritParams font_info
#'
#' @return A data frame with a `path` and `index` column giving fallback for the
#' specified string and font combinations
#'
#' @export
#'
#' @examples
#' font_fallback("\U0001f604") # Smile emoji
#'
font_fallback <- function(
  string,
  family = '',
  italic = FALSE,
  weight = "normal",
  width = "undefined",
  path = NULL,
  index = 0,
  variation = font_variation(),
  bold = deprecated()
) {
  if (lifecycle::is_present(bold)) {
    lifecycle::deprecate_soft(
      "1.2.4",
      "font_fallback(bold)",
      "font_fallback(weight)"
    )
    weight <- ifelse(bold, "bold", "normal")
  }
  if (is_font_variation(variation)) variation <- list(variation)
  full_length <- length(string)
  if (is.null(path)) {
    fonts <- match_fonts(
      family = rep_len_default(family, full_length, ''),
      italic = rep_len_default(italic, full_length, FALSE),
      weight = rep_len_default(weight, full_length, "normal"),
      width = rep_len_default(width, full_length, "undefined")
    )
    path <- fonts$path
    index <- fonts$index
  } else {
    full_length <- max(length(path), length(index), full_length)
    if (!all(c(length(path), length(index)) == 1)) {
      path <- rep_len(path, full_length)
      index <- rep_len(index, full_length)
    }
  }
  if (length(string) != 1) string <- rep_len(string, full_length)
  variation <- rep_len(variation, full_length)
  if (!all(file.exists(path)))
    stop("path must point to a valid file", call. = FALSE)
  get_fallback_c(path, as.integer(index), as.character(string), variation)
}

#' Register font collections as families
#'
#' By design, systemfonts searches the fonts installed natively on the system.
#' It is possible, however, to register other fonts from e.g. font packages or
#' local font files, that will get searched before searching any installed
#' fonts. You can always get an overview over all registered fonts with the
#' `registry_fonts()` function that works as a registry focused analogue to
#' [system_fonts()]. If you wish to clear out the registry, you can either
#' restart the R session or call `clear_registry()`.
#'
#' @inheritSection match_fonts Font matching
#'
#' @param name The name the collection will be known under (i.e. *family*)
#' @param plain,bold,italic,bolditalic Fontfiles for the different faces of the
#' collection. can either be a filepath or a list containing a filepath and an
#' index (only for font files containing multiple fonts). If not given it will
#' default to the `plain` specification.
#' @param features A [`font_feature`] object describing the specific OpenType
#' font features to turn on for the registered font.
#'
#' @return `register_font()` and `clear_registry()` returns `NULL` invisibly.
#' `registry_fonts()` returns a data table in the same style as [system_fonts()]
#' though less detailed and not based on information in the font file.
#'
#' @details
#' `register_font` also makes it possible to use system fonts with traits that
#' is not covered by the graphic engine in R. In plotting operations it is only
#' possible to specify a family name and whether or not the font should be bold
#' and/or italic. There are numerous fonts that will never get matched to this,
#' especially because bold is only one of many weights.
#'
#' Apart from granting a way to use new varieties of fonts, font registration
#' also allows you to override the default `sans`, `serif`, and `mono` mappings,
#' simply by registering a collection to the relevant default name. As
#' registered fonts are searched first it will take precedence over the default.
#'
#' @export
#'
#' @examplesIf nrow(system_fonts()) >= 4
#' # Create a random font collection
#' fonts <- system_fonts()
#' random_fonts <- sample(nrow(fonts), 4)
#' register_font(
#'   'random',
#'   plain = list(fonts$path[random_fonts[1]], fonts$index[random_fonts[1]]),
#'   bold = list(fonts$path[random_fonts[2]], fonts$index[random_fonts[2]]),
#'   italic = list(fonts$path[random_fonts[3]], fonts$index[random_fonts[3]]),
#'   bolditalic = list(fonts$path[random_fonts[4]], fonts$index[random_fonts[4]])
#' )
#'
#' # Look at your creation
#' registry_fonts()
#'
#' # Reset
#' clear_registry()
#'
register_font <- function(
  name,
  plain,
  bold = plain,
  italic = plain,
  bolditalic = plain,
  features = font_feature()
) {
  if (name %in% system_fonts()$family) {
    stop("A system font called `", name, "` already exists", call. = FALSE)
  }
  if (is.character(plain)) {
    plain <- list(plain, 0)
  }
  if (is.character(bold)) {
    bold <- list(bold, 0)
  }
  if (is.character(italic)) {
    italic <- list(italic, 0)
  }
  if (is.character(bolditalic)) {
    bolditalic <- list(bolditalic, 0)
  }
  files <- c(plain[[1]], bold[[1]], italic[[1]], bolditalic[[1]])
  indices <- c(plain[[2]], bold[[2]], italic[[2]], bolditalic[[2]])
  if (!all(file.exists(files))) {
    stop(name, " refers to non-existing font file(s)", call. = FALSE)
  }

  register_font_c(
    as.character(name),
    as.character(files),
    as.integer(indices),
    features[[1]],
    features[[2]]
  )
}
#' @rdname register_font
#' @export
registry_fonts <- function() {
  registry_fonts_c()
}
#' @rdname register_font
#' @export
clear_registry <- function() {
  clear_registry_c()
}

#' Register a font as a variant as an existing one
#'
#' This function is a wrapper around [register_font()] that allows you to easily
#' create variants of existing system fonts, e.g. to target different weights
#' and/or widths, or for attaching OpenType features to a font.
#'
#' @inheritSection match_fonts Font matching
#'
#' @param name The new family name the variant should respond to
#' @param family The name of an existing font family that this is a variant of
#' @param weight One or two of `"thin"`, `"ultralight"`, `"light"`, `"normal"`,
#' `"medium"`, `"semibold"`, `"bold"`, `"ultrabold"`, or `"heavy"`. If one is
#' given it sets the weight for the whole variant. If two is given the first
#' one defines the plain weight and the second the bold weight. If `NULL` then
#' the variants of the given family closest to `"normal"` and `"bold"` will be
#' chosen.
#' @param width One of `"ultracondensed"`, `"extracondensed"`, `"condensed"`,
#' `"semicondensed"`, `"normal"`, `"semiexpanded"`, `"expanded"`,
#' `"extraexpanded"`, or `"ultraexpanded"` giving the width of the variant. If
#' `NULL` then the width closest to `"normal"` will be chosen.
#' @param features A [`font_feature`] object describing the specific OpenType
#' font features to turn on for the registered font variant.
#'
#' @export
#'
#' @examples
#' # Get the default "sans" family
#' sans <- match_fonts("sans")$path
#' sans <- system_fonts()$family[system_fonts()$path == sans][1]
#'
#' # Register a variant of it:
#' register_variant(
#'   "sans_ligature",
#'   sans,
#'   features = font_feature(ligatures = "discretionary")
#' )
#'
#' registry_fonts()
#'
#' # clean up
#' clear_registry()
#'
register_variant <- function(
  name,
  family,
  weight = NULL,
  width = NULL,
  features = font_feature()
) {
  sys_fonts <- system_fonts()
  sys_fonts <- sys_fonts[
    grepl(tolower(family), tolower(sys_fonts$family)),
    ,
    drop = FALSE
  ]
  if (!is.null(width)) {
    sys_fonts <- sys_fonts[sys_fonts$width == tolower(width), , drop = FALSE]
  }
  if (!is.null(weight)) {
    sys_fonts <- sys_fonts[
      sys_fonts$weight %in% tolower(weight),
      ,
      drop = FALSE
    ]
  }
  if (nrow(sys_fonts) == 0) {
    stop(
      "No font with the given family name and weight/width settings available",
      call. = FALSE
    )
  }
  if (any(tolower(family) == tolower(sys_fonts$family))) {
    sys_fonts <- sys_fonts[
      tolower(family) == tolower(sys_fonts$family),
      ,
      drop = FALSE
    ]
  }
  if (is.null(width)) {
    width <- sys_fonts$width[which.min(abs(as.integer(sys_fonts$width) - 5))]
    sys_fonts <- sys_fonts[sys_fonts$width == tolower(width), , drop = FALSE]
  }
  if (is.null(weight)) {
    normal <- which.min(abs(as.integer(sys_fonts$weight) - 4))
    bold <- which.min(abs(as.integer(sys_fonts$weight) - 7))
    weight <- sys_fonts$weight[unique(c(normal, bold))]
  }
  plain <- sys_fonts[
    which(sys_fonts$weight == weight[1] & !sys_fonts$italic),
    ,
    drop = FALSE
  ]
  bold <- if (length(weight) == 2) {
    sys_fonts[
      which(sys_fonts$weight == weight[2] & !sys_fonts$italic),
      ,
      drop = FALSE
    ]
  } else {
    plain
  }
  italic <- sys_fonts[
    which(sys_fonts$weight == weight[1] & sys_fonts$italic),
    ,
    drop = FALSE
  ]
  bolditalic <- if (length(weight) == 2) {
    sys_fonts[
      which(sys_fonts$weight == weight[2] & sys_fonts$italic),
      ,
      drop = FALSE
    ]
  } else {
    italic
  }
  if (nrow(plain) == 0) {
    plain <- italic
  }
  if (nrow(bold) == 0) {
    bold <- plain
  }
  if (nrow(italic) == 0) {
    italic <- plain
  }
  if (nrow(bolditalic) == 0) {
    bolditalic <- if (length(weight) == 2) bold else italic
  }
  register_font(
    name,
    as.list(plain[1, c('path', 'index')]),
    as.list(bold[1, c('path', 'index')]),
    as.list(italic[1, c('path', 'index')]),
    as.list(bolditalic[1, c('path', 'index')]),
    features
  )
}

#' Add local font files to the search path
#'
#' systemfonts is mainly about getting system native access to the fonts
#' installed on the OS you are executing the code on. However, you may want to
#' access fonts without doing a full installation, either because you want your
#' project to be reproducible on all systems, because you don't have
#' administrator privileges on the system, or for a different reason entirely.
#' `add_fonts()` provide a way to side load font files so that they are found
#' during font matching. The function differs from [register_font()] and
#' [register_variant()] in that they add the font file as-is using the family
#' name etc that are provided by the font. `scan_local_fonts()` is run when
#' systemfonts is loaded and will automatically add font files stored in
#' `./fonts` (project local) and `~/fonts` (user local).
#'
#' @inheritSection match_fonts Font matching
#'
#' @param files A character vector of font file paths or urls to add
#'
#' @return This function is called for its sideeffects
#'
#' @export
#'
#' @examples
#' # example code
#' empty_font <- system.file("unfont.ttf", package = "systemfonts")
#'
#' add_fonts(empty_font)
#'
#' clear_local_fonts()
#'
add_fonts <- function(files) {
  if (!is.character(files)) {
    stop("`files` must be a character vector")
  }
  urls <- grepl("^https?://", files)
  if (any(urls)) {
    dest <- vapply(which(urls), function(i) tempfile(), character(1))
    success <- utils::download.file(files[urls], dest, method = "libcurl")
    if (success != 0) {
      stop(
        "Download of font files failed with the libcurl error ",
        success,
        call. = FALSE
      )
    }
    files[urls] <- dest
  }
  do_exist <- file.exists(files)
  if (!all(do_exist)) {
    warning(
      "all elements in `files` must be paths to existing files. Ignoring ",
      paste0(files[!do_exist], collapse = ", ")
    )
    files <- files[do_exist]
  }
  if (length(files) > 0) {
    add_local_fonts(vapply(files, normalizePath, character(1)))
  }
  invisible(NULL)
}

#' @rdname add_fonts
#' @export
#'
scan_local_fonts <- function() {
  files <- unique(c(
    list.files(
      "./fonts",
      all.files = TRUE,
      full.names = TRUE,
      recursive = TRUE
    ),
    list.files("~/fonts", all.files = TRUE, full.names = TRUE, recursive = TRUE)
  ))
  add_fonts(files)
}
#' @rdname add_fonts
#' @export
#'
clear_local_fonts <- function() {
  message(
    "Run `scan_local_fonts()` in order to re-add automatically added fonts"
  )
  clear_local_fonts_c()
}

#' Get the outline of glyphs
#'
#' This function allows you to retrieve the outline of glyphs as polygon
#' coordinates. The glyphs are given as indexes into a font file and not as
#' characters allowing you to retrieve outlines for glyphs that doesn't have a
#' character counterpoint. Glyphs that are given as bitmaps are ignored.
#'
#' @param glyph The index of the glyph in the font file
#' @param path The path to the font file encoding the glyph
#' @param index The index of the font in the font file
#' @param size The size of the font in big points (1/72 inch)
#' @param variation A `font_variation` object or a list of them to control
#' variable fonts
#' @param tolerance The deviation tolerance for decomposing bezier curves of the
#' glyph. Given in the same unit as size. Smaller values give more detailed
#' polygons
#' @param verbose Should font and glyph loading errors be reported as warnings
#'
#' @return A data frame giving the outlines of the glyphs provide in `glyph`. It
#' contains the columns `glyph` pointing to the element in the input it relates
#' to, `contour` enumerating the contours the glyph consists of, and `x` and `y`
#' giving the coordinates in big points
#'
#' @export
#'
#' @examples
#' # Get the shape of s in the default font
#' font <- font_info()
#' glyph <- glyph_info("s", path = font$path, index = font$index)
#'
#' s <- glyph_outline(glyph$index, font$path, font$index, size = 150)
#'
#' plot(s$x, s$y, type = 'l')
#'
glyph_outline <- function(
  glyph,
  path,
  index = 0,
  size = 12,
  tolerance = 0.2,
  variation = font_variation(),
  verbose = FALSE
) {
  if (is_font_variation(variation)) variation <- list(variation)
  n_glyphs <- length(glyph)
  glyph <- as.integer(glyph)
  path <- rep_len(as.character(path), n_glyphs)
  index <- rep_len_default(as.integer(index), n_glyphs, 0L)
  size <- rep_len_default(as.numeric(size), n_glyphs, 12)
  variation <- rep_len(variation, n_glyphs)
  tolerance <- as.numeric(tolerance)

  get_glyph_outlines(
    glyph,
    path,
    index,
    size,
    variation,
    tolerance,
    as.logical(verbose)
  )
}

#' Render glyphs to raster image
#'
#' Not all glyphs are encoded as vector outlines (emojis often not). Even for
#' fonts that provide an outline you might be interested in a raster version.
#' This function gives you just that. It converts a glyph into an optimized
#' raster object that can be plotted with e.g. [graphics::rasterImage()] or
#' [grid::grid.raster()]. For convenience, you can also use
#' [glyph_raster_grob()] for plotting the result.
#'
#' @inheritParams glyph_outline
#' @param res The resolution to render the glyphs to
#' @param col The color of the glyph assuming the glyph doesn't have a native
#' coloring
#'
#' @return A list of nativeRaster objects (or `NULL` if it failed to render a
#' given glyph). The nativeRasters have additional attributes attached. `"size"`
#' will give the size of the glyph in big points and `"offset"` will give the
#' location of the top-left corner of the raster with respect to where it should
#' be rendered.
#'
#' @export
#'
#' @examples
#' font <- font_info()
#' glyph <- glyph_info("R", path = font$path, index = font$index)
#'
#' R <- glyph_raster(glyph$index, font$path, font$index, size = 150)
#'
#' plot.new()
#' plot.window(c(0,150), c(0, 150), asp = 1)
#' rasterImage(R[[1]], 0, 0, attr(R[[1]], "size")[2], attr(R[[1]], "size")[1])
#'
glyph_raster <- function(
  glyph,
  path,
  index = 0,
  size = 12,
  res = 300,
  variation = font_variation(),
  col = "black",
  verbose = FALSE
) {
  if (is_font_variation(variation)) variation <- list(variation)
  n_glyphs <- length(glyph)

  if (all(col == "black" | col == "#000000")) {
    col <- rep_len(-16777216L, n_glyphs)
  } else {
    if (!requireNamespace("farver", quietly = TRUE)) {
      stop("The farver package is required to tint glyphs with a color")
    }
    col <- rep_len(farver::encode_native(col), n_glyphs)
  }

  glyph <- as.integer(glyph)
  path <- rep_len(as.character(path), n_glyphs)
  index <- rep_len_default(as.integer(index), n_glyphs, 0L)
  size <- rep_len_default(as.numeric(size), n_glyphs, 12)
  res <- rep_len_default(as.numeric(res), n_glyphs, 300)
  variation <- rep_len(variation, n_glyphs)

  get_glyph_bitmap(glyph, path, index, size, res, variation, col, verbose)
}

#' Convert an extracted glyph raster to a grob
#'
#' This is a convenience function that helps in creating [rasterGrob] with the
#' correct settings for the glyph. It takes inot account the sizing and offset
#' returned by [glyph_raster()] and allows you to only consider the baseline
#' position of the glyph.
#'
#' @param glyph The nativeRaster object returned as one of the elements by
#' [glyph_raster()]
#' @param x,y The baseline location of the glyph
#' @inheritDotParams grid::rasterGrob -x -y
#' @inheritParams grid::rasterGrob
#'
#' @return A rasterGrob object
#'
#' @export
#'
#' @examples
#' font <- font_info()
#' glyph <- glyph_info("R", path = font$path, index = font$index)
#'
#' R <- glyph_raster(glyph$index, font$path, font$index, size = 150)
#'
#' grob <- glyph_raster_grob(R[[1]], 50, 50)
#'
#' grid::grid.newpage()
#' # Mark the baseline location
#' grid::grid.points(50, 50, default.units = "bigpts")
#' # Draw the glyph
#' grid::grid.draw(grob)
#'
glyph_raster_grob <- function(glyph, x, y, ..., default.units = "bigpts") {
  if (length(glyph) == 0) return(grid::nullGrob())
  if (!grid::is.unit(x)) x <- grid::unit(x, default.units)
  if (!grid::is.unit(y)) y <- grid::unit(y, default.units)
  size <- attr(glyph, "size")
  offset <- attr(glyph, "offset")
  grid::rasterGrob(
    glyph,
    x = x + grid::unit(offset[2], "bigpts"),
    y = y + grid::unit(offset[1], "bigpts"),
    width = grid::unit(size[2], "bigpts"),
    height = grid::unit(size[1], "bigpts"),
    just = c("left", "top"),
    ...
  )
}

#' Create a visual representation of what the various glyph stats mean
#'
#' This function helps you understand the concepts of width, height, bearing,
#' and advance by annotating a glyph with the various measures
#'
#' @param glyph The character to plot
#' @inheritParams glyph_info
#'
#' @return This function is called for its side effects
#'
#' @export
#'
#' @examples
#' plot_glyph_stats("g")
#'
plot_glyph_stats <- function(
  glyph,
  family = '',
  italic = FALSE,
  weight = "normal",
  width = "undefined",
  size = 12,
  res = 72,
  variation = font_variation(),
  path = NULL,
  index = 0
) {
  font <- font_info(
    family = family,
    italic = italic,
    weight = weight,
    width = width,
    path = path,
    index = index
  )
  info <- glyph_info(
    glyphs = glyph,
    family = family,
    italic = italic,
    weight = weight,
    width = width,
    size = size,
    res = res,
    variation = variation,
    path = path,
    index = index
  )
  outline <- glyph_outline(
    glyph = info$index[1],
    path = font$path[1],
    index = font$index[1],
    size = size,
    variation = variation,
    tolerance = 0.1
  )
  x_scale <- c(min(0, info$x_bearing), info$x_advance)
  y_scale <- c(min(0, info$bbox[[1]][3]), info$bbox[[1]][4])
  max_dim <- max(diff(x_scale), diff(y_scale))
  gutter <- max_dim * 0.1 * c(-1, 1)
  x_scale <- x_scale + gutter
  y_scale <- y_scale + gutter
  max_dim <- max_dim * 1.2
  vp <- grid::viewport(
    width = grid::unit(diff(x_scale) / max_dim, "snpc"),
    height = grid::unit(diff(y_scale) / max_dim, "snpc"),
    xscale = x_scale,
    yscale = y_scale,
    clip = "off"
  )
  bbox <- grid::rectGrob(
    x = info$bbox[[1]][1],
    y = info$bbox[[1]][3],
    width = info$bbox[[1]][2] - info$bbox[[1]][1],
    height = info$bbox[[1]][4] - info$bbox[[1]][3],
    hjust = 0,
    vjust = 0,
    gp = grid::gpar(col = NA, fill = "grey90"),
    default.units = "native",
    vp = vp
  )
  glyph <- grid::pathGrob(
    outline$x,
    outline$y,
    id = outline$contour,
    gp = grid::gpar(col = "black", fill = "white"),
    default.units = "native",
    vp = vp
  )
  coord <- grid::segmentsGrob(
    x0 = grid::unit(c(0, 0), c("npc", "native")),
    y0 = grid::unit(c(0, 0), c("native", "npc")),
    x1 = grid::unit(c(1, 0), c("npc", "native")),
    y1 = grid::unit(c(0, 1), c("native", "npc")),
    vp = vp,
    gp = grid::gpar(lty = 3)
  )
  origin <- grid::pointsGrob(
    0,
    0,
    vp = vp,
    pch = 19,
    size = grid::unit(0.03, "snpc")
  )
  arrows <- grid::segmentsGrob(
    x0 = c(
      info$x_bearing,
      info$bbox[[1]][2] + max_dim * 0.02,
      info$bbox[[1]][1] - max_dim * 0.02,
      0,
      0
    ),
    y0 = c(
      info$y_bearing + max_dim * 0.02,
      info$y_bearing,
      0,
      info$y_bearing + max_dim * 0.05,
      info$bbox[[1]][3] - max_dim * 0.02
    ),
    x1 = c(
      info$x_bearing + info$width,
      info$bbox[[1]][2] + max_dim * 0.02,
      info$bbox[[1]][1] - max_dim * 0.02,
      info$x_bearing,
      info$x_advance
    ),
    y1 = c(
      info$y_bearing + max_dim * 0.02,
      info$y_bearing - info$height,
      info$y_bearing,
      info$y_bearing + max_dim * 0.05,
      info$bbox[[1]][3] - max_dim * 0.02
    ),
    default.units = "native",
    arrow = grid::arrow(
      length = grid::unit(0.015, "snpc"),
      ends = c("both", "both", "last", "last", "last")
    ),
    vp = vp
  )
  labels <- grid::textGrob(
    label = c(
      "width",
      "height",
      "x bearing",
      "y bearing",
      "x advance",
      "(0,0)"
    ),
    x = c(
      mean(info$bbox[[1]][1:2]),
      info$bbox[[1]][2] + max_dim * 0.03,
      min(0, info$x_bearing) - max_dim * 0.01,
      info$bbox[[1]][1] - max_dim * 0.03,
      info$x_advance / 2,
      -max_dim * 0.01
    ),
    y = c(
      info$y_bearing + max_dim * 0.03,
      mean(info$bbox[[1]][3:4]),
      info$y_bearing + max_dim * 0.05,
      info$y_bearing / 2,
      info$bbox[[1]][3] - max_dim * 0.03,
      -max_dim * 0.01
    ),
    hjust = c(0.5, 0, 1, 1, 0.5, 1),
    vjust = c(0, 0.5, 0.5, 0.5, 1, 1),
    default.units = "native",
    vp = vp
  )
  grid::grid.newpage()
  grid::grid.draw(grid::gList(bbox, coord, origin, glyph, arrows, labels))
}

#' List all fonts installed on your system
#'
#' @return A data frame with a row for each font and various information in each
#' column
#'
#' @export
#'
#' @examples
#' # See all monospace fonts
#' fonts <- system_fonts()
#' fonts[fonts$monospace, ]
#'
system_fonts <- function() {
  system_fonts_c()
}

#' Reset the system font cache
#'
#' Building the list of system fonts is time consuming and is therefore cached.
#' This, in turn, means that changes to the system fonts (i.e. installing new
#' fonts), will not propagate to systemfonts. The solution is to reset the
#' cache, which will result in the next call to e.g. [match_fonts()] will
#' trigger a rebuild of the cache.
#'
#' @export
#'
#' @examples
#' all_fonts <- system_fonts()
#'
#' ##-- Install a new font on the system --##
#'
#' all_fonts_new <- system_fonts()
#'
#' ## all_fonts_new will be equal to all_fonts
#'
#' reset_font_cache()
#'
#' all_fonts_new <- system_fonts()
#'
#' ## all_fonts_new will now contain the new font
#'
reset_font_cache <- function() {
  reset_font_cache_c()
}

context("Font listing")

test_that("System fonts can be listed", {
  fonts <- system_fonts()
  expect_is(fonts, 'data.frame')
  expect_gt(nrow(fonts), 0)
  expect_named(
    fonts,
    c(
      "path",
      "index",
      "name",
      "family",
      "style",
      "weight",
      "width",
      "italic",
      "monospace",
      "variable"
    )
  )
})

#' Define axis coordinates for variable fonts
#'
#' Variable fonts is a technology that allows font designers to encode the full,
#' continuous font space of a typeface into a single font file and then have the
#' user set the coordinates of the variable axes to define the font. So, rather
#' than having a font file for bold, bold + italic, thin, and thin + italic,
#' etc. it is all encoded in a single file with a continuous range of all axes
#' (e.g. weight doesn't have to be one of the 9 standard weights but can be
#' anything in between). There are 5 standard axes that fonts can use, but font
#' designers are free to define their own completely arbitrary axes as well. You
#' can use [font_info()] to see which axes a font defines along with their value
#' range and default setting. Values given as `font_variation()` will always win
#' over the conventional setting *if* the axis is present in the font. For
#' example, setting `weight = "bold"` along with
#' `variation = font_variation(weight = 650)` will eventually request a weight
#' of `650` (helfway between semibold and bold), assuming the weight-axis is
#' present in the font. For clarity however, it is advised that `font_variation()`
#' is only used for axes that can otherwise not be accessed by "top-level"
#' arguments.
#'
#' @param italic Value between 0 and 1. Usually treated as a boolean rather than
#' a continuous axis
#' @param weight Usually a value between 100 and 900 though fonts can limit or
#' expand the supported range. Weight names are also allowed (see
#' [as_font_weight()])
#' @param width Usually a value between 1 and 9 though fonts can limit or
#' expand the supported range. Width names are also allowed (see
#' [as_font_width()])
#' @param slant The angular slant of the font, usually between -90 and 90
#' (negative values tilt in the "standard" italic way)
#' @param optical_sizing Stroke thickness compensation for the glyphs. During
#' rendering the thickness of the stroke is usually increased when the font is
#' set at small sizes to increase readability. Set this to the size of the font
#' to get the "expected" look at that size.
#' @param ... Further axes and coordinate settings
#'
#' @return A `font_variation` object with coordinates for the provided axes
#'
#' @note systemfonts uses a scale of width values ranging from 1-9 while the
#' width axis uses a different scale (0.5 - 2.0) going from half as wide to
#' twice as wide as "normal". When using the `width` argument the coordinate
#' values is automatically converted. If you set the value based on the width
#' tag (`wdth`) then no conversion will happen.
#'
#'
#' @export
#'
font_variation <- function(
  italic = NULL,
  weight = NULL,
  width = NULL,
  slant = NULL,
  optical_sizing = NULL,
  ...
) {
  variations <- list(...)
  if (
    length(variations) == 0 &&
      is.null(italic) &&
      is.null(weight) &&
      is.null(width) &&
      is.null(slant) &&
      is.null(optical_sizing)
  ) {
    return(structure(
      list(axis = integer(0), value = integer(0)),
      class = "font_variation"
    ))
  }
  if (is.null(variations$ital) && !is.null(italic)) {
    variations$ital <- as.numeric(italic)
  }
  if (is.null(variations$wght) && !is.null(weight)) {
    variations$wght <- as_font_weight(weight)
    if (variations$wght == 0) variations$wght <- NULL
  }
  if (is.null(variations$wdth) && !is.null(width)) {
    variations$wdth <- as_font_width(width)
    if (variations$wdth == 0) {
      variations$wdth <- NULL
    } else {
      variations$wdth <- 2^((variations$wdth - 5) / 4)
    }
  }
  if (is.null(variations$slnt) && !is.null(slant)) {
    variations$slnt <- as.numeric(slant)
  }
  # TODO: Support "auto" (i.e. set it to the size of the font)
  if (is.null(variations$opsz) && !is.null(optical_sizing)) {
    variations$optz <- as.numeric(optical_sizing)
  }
  variations <- variations[lengths(variations) != 0]
  axes <- names(variations) %||% character(0)
  values <- vapply(variations, identity, numeric(1))
  structure(
    list(
      axis = axes_to_tags(axes),
      value = values_to_fixed(values)
    ),
    class = "font_variation"
  )
}
is_font_variation <- function(x) inherits(x, "font_variation")
#' @export
print.font_variation <- function(x, ...) {
  if (length(x) == 0) {
    cat("An empty font_variation object\n")
  } else {
    cat("A list of font variation axis coordinates\n")
    cat(paste(paste0("- ", axes(x), ": ", coords(x)), collapse = "\n"))
  }
  invisible(x)
}
#' @export
format.font_variation <- function(x, ...) {
  if (length(x) == 0) {
    return("<empty>")
  }
  paste(axes(x), ': ', coords(x), collapse = '; ', sep = '')
}
combine_variations <- function(a, b) {
  a <- unclass(a)
  b <- unclass(b)
  new <- setdiff(b$axis, a$axis)
  current <- intersect(b$axis, a$axis)
  a$axis <- c(a$axis, new)
  a$value[match(b$axis, a$axis)] <- b$value
  `class<-`(a, "font_variation")
}
add_standard_to_variations <- function(variations, italic, weight, width) {
  mapply(
    function(orig_var, italic, weight, width) {
      combine_variations(
        font_variation(italic = italic, weight = weight, width = width),
        orig_var
      )
    },
    orig_var = variations,
    italic = italic,
    weight = weight,
    width = width,
    SIMPLIFY = FALSE
  )
}

axes <- function(x) {
  tags_to_axes(.subset2(x, "axis"))
}
coords <- function(x) {
  fixed_to_values(.subset2(x, "value"))
}

#' @export
length.font_variation <- function(x) {
  length(.subset2(x, "axis"))
}
#' @export
names.font_variation <- function(x) {
  axes(x)
}
#' @export
`[[.font_variation` <- function(x, i, ...) {
  if (is.character(i)) {
    coords(x)[[match(i, axes(x))]]
  } else {
    coords(x)[[i]]
  }
}
#' @export
`[[<-.font_variation` <- function(x, i, value) {
  axes <- .subset2(x, "axis")
  values <- .subset2(x, "value")
  if (is.character(i)) {
    if (i == "width") {
      i <- "wdth"
      value <- 2^((value - 5) / 4)
    } else if (i == "weight") {
      i <- "wght"
    } else if (i == "italic") {
      i <- "ital"
    } else if (i == "slant") {
      i <- "slnt"
    } else if (i == "optical_sizing") {
      i <- "optz"
    }
    i <- axes_to_tags(i)
    ind <- match(i, axes)
  } else if (i > length(axes)) {
    stop("subscript out of bounds", call. = FALSE)
  } else {
    ind <- i
  }
  value <- values_to_fixed(value)
  if (is.na(ind)) {
    axes <- c(axes, i)
    values <- c(values, value)
  } else {
    values[ind] <- value
  }
  structure(
    list(axis = axes, value = values),
    class = "font_variation"
  )
}
#' @export
`$.font_variation` <- function(x, name) {
  x[[name]]
}
#' @export
`$<-.font_variation` <- function(x, name, value) {
  `[[<-`(x, name, value)
}
#' @export
`[.font_variation` <- function(x, i, ...) {
  if (is.character(i)) {
    coords(x)[match(i, axes(x))]
  } else {
    coords(x)[i]
  }
}
#' @export
`[<-.font_variation` <- function(x, i, value) {
  if (length(value) != length(i)) {
    stop("`i` and `value` must be the same length")
  }
  for (j in seq_along(i)) {
    x[[i[[j]]]] <- value[[j]]
  }
  x
}

if (!file.exists("../windows/harfbuzz/include/harfbuzz/hb.h")) {
  unlink("../windows", recursive = TRUE)
  url <- if (grepl("aarch", R.version$platform)) {
    "https://github.com/r-windows/bundles/releases/download/harfbuzz-8.2.1/harfbuzz-8.2.1-clang-aarch64.tar.xz"
  } else if (grepl("clang", Sys.getenv('R_COMPILED_BY'))) {
    "https://github.com/r-windows/bundles/releases/download/harfbuzz-8.2.1/harfbuzz-8.2.1-clang-x86_64.tar.xz"
  } else if (getRversion() >= "4.2") {
    "https://github.com/r-windows/bundles/releases/download/harfbuzz-8.2.1/harfbuzz-8.2.1-ucrt-x86_64.tar.xz"
  } else {
    "https://github.com/rwinlib/harfbuzz/archive/v2.7.4.tar.gz"
  }
  download.file(url, basename(url), quiet = TRUE)
  dir.create("../windows", showWarnings = FALSE)
  untar(basename(url), exdir = "../windows", tar = 'internal')
  unlink(basename(url))
  setwd("../windows")
  file.rename(list.files(), 'freetype2')
}

#' Get string widths as measured by the current device
#'
#' For certain composition tasks it is beneficial to get the width of a string
#' as interpreted by the device that is going to plot it. grid provides this
#' through construction of a `textGrob` and then converting the corresponding
#' grob width to e.g. cm, but this comes with a huge overhead.
#' `string_widths_dev()` provides direct, vectorised, access to the graphic
#' device for as high performance as possible.
#'
#' @param strings A character vector of strings to measure
#' @param family The font families to use. Will get recycled
#' @param face The font faces to use. Will get recycled
#' @param size The font size to use. Will get recycled
#' @param cex The cex multiplier to use. Will get recycled
#' @param unit The unit to return the width in. Either `"cm"`, `"inches"`,
#' `"device"`, or `"relative"`
#'
#' @return A numeric vector with the width of each of the strings given in
#' `strings` in the unit given in `unit`
#'
#' @export
#'
#' @family device metrics
#'
#' @examples
#' # Get the widths as measured in cm (default)
#' string_widths_dev(c('a string', 'an even longer string'))
#'
string_widths_dev <- function(
  strings,
  family = '',
  face = 1,
  size = 12,
  cex = 1,
  unit = 'cm'
) {
  unit <- match.arg(unit, possible_units)
  unit <- match(unit, possible_units) - 1L
  n_total <- length(strings)
  if (length(family) != 1) family <- rep_len_default(family, n_total, '')
  if (any(c(length(face), length(size), length(cex)) != 1)) {
    face <- rep_len_default(face, n_total, 1)
    size <- rep_len_default(size, n_total, 12)
    cex <- rep_len_default(cex, n_total, 1)
  }
  dev_string_widths_c(
    as.character(strings),
    as.character(family),
    as.integer(face),
    as.numeric(size),
    as.numeric(cex),
    unit
  )
}
#' Get string metrics as measured by the current device
#'
#' This function is much like [string_widths_dev()] but also returns the ascent
#' and descent of the string making it possible to construct a tight bounding
#' box around the string.
#'
#' @inheritParams string_widths_dev
#'
#' @return A data.frame with `width`, `ascent`, and `descent` columns giving the
#' metrics in the requested unit.
#'
#' @family device metrics
#'
#' @export
#'
#' @examples
#' # Get the metrics as measured in cm (default)
#' string_metrics_dev(c('some text', 'a string with descenders'))
#'
string_metrics_dev <- function(
  strings,
  family = '',
  face = 1,
  size = 12,
  cex = 1,
  unit = 'cm'
) {
  unit <- match.arg(unit, possible_units)
  unit <- match(unit, possible_units) - 1L
  n_total <- length(strings)
  if (length(family) != 1) family <- rep_len_default(family, n_total, '')
  if (any(c(length(face), length(size), length(cex)) != 1)) {
    face <- rep_len_default(face, n_total, 1)
    size <- rep_len_default(size, n_total, 12)
    cex <- rep_len_default(cex, n_total, 1)
  }
  dev_string_metrics_c(
    as.character(strings),
    as.character(family),
    as.integer(face),
    as.numeric(size),
    as.numeric(cex),
    unit
  )
}

# Order important. Will get converted to 0-indexed unit identity for C code
possible_units <- c('cm', 'inches', 'device', 'relative')

#' Query font-specific information
#'
#' Get general information about a font, relative to a given size. Size specific
#' measures will be returned in pixel units. The function is vectorised to the
#' length of the longest argument.
#'
#' @inheritParams match_font
#' @param size The pointsize of the font to use for size related measures
#' @param res The ppi of the size related measures
#' @param variation A `font_variation` object or a list of them to control
#' variable fonts
#' @param path,index path and index of a font file to circumvent lookup based on
#' family and style
#' @param bold `r lifecycle::badge("deprecated")` Use `weight = "bold"` instead
#'
#' @return
#' A data.frame giving info on the requested font + size combinations. The
#' data.frame will contain the following columns:
#'
#' \describe{
#'   \item{path}{The path to the font file}
#'   \item{index}{The 0-based index of the font in the fontfile}
#'   \item{family}{The family name of the font}
#'   \item{style}{The style name of the font}
#'   \item{name}{The name of the font, if present, otherwise the family name}
#'   \item{italic}{A logical giving if the font is italic}
#'   \item{bold}{A logical giving if the font is bold}
#'   \item{monospace}{A logical giving if the font is monospace}
#'   \item{weight}{A factor giving the weight of the font}
#'   \item{width}{A factor giving the width of the font}
#'   \item{kerning}{A logical giving if the font supports kerning}
#'   \item{color}{A logical giving if the font has color glyphs}
#'   \item{scalable}{A logical giving if the font is scalable}
#'   \item{vertical}{A logical giving if the font is vertical}
#'   \item{n_glyphs}{The number of glyphs in the font}
#'   \item{n_sizes}{The number of predefined sizes in the font}
#'   \item{n_charmaps}{The number of character mappings in the font file}
#'   \item{bbox}{A bounding box large enough to contain any of the glyphs in the font}
#'   \item{max_ascend}{The maximum ascend of the tallest glyph in the font}
#'   \item{max_descent}{The maximum descend of the most descending glyph in the font}
#'   \item{max_advance_width}{The maximum horizontal advance a glyph can make}
#'   \item{max_advance_height}{The maximum vertical advance a glyph can make}
#'   \item{lineheight}{The height of a single line of text in the font}
#'   \item{underline_pos}{The position of a potential underlining segment}
#'   \item{underline_size}{The width the the underline}
#' }
#'
#' @export
#'
#' @examples
#' font_info('serif')
#'
#' # Avoid lookup if font file is already known
#' sans <- match_fonts('sans')
#' font_info(path = sans$path, index = sans$index)
#'
font_info <- function(
  family = '',
  italic = FALSE,
  weight = "normal",
  width = "undefined",
  size = 12,
  res = 72,
  path = NULL,
  index = 0,
  variation = font_variation(),
  bold = deprecated()
) {
  if (is_font_variation(variation)) variation <- list(variation)
  full_length <- max(
    length(size),
    length(res),
    length(italic),
    length(weight),
    length(width),
    length(variation)
  )
  if (is.null(path)) {
    if (lifecycle::is_present(bold)) {
      lifecycle::deprecate_soft("1.2.4", "font_info(bold)", "font_info(weight)")
      weight <- ifelse(bold, "bold", "normal")
    }
    full_length <- max(
      length(family),
      full_length
    )
    italic <- rep_len_default(italic, full_length, FALSE)
    weight <- rep_len_default(weight, full_length, "normal")
    width <- rep_len_default(width, full_length, "undefined")
    fonts <- match_fonts(
      family = rep_len_default(family, full_length, ''),
      italic = italic,
      weight = weight,
      width = width
    )
    path <- fonts$path
    index <- fonts$index
  } else {
    full_length <- max(length(path), length(index), full_length)
    italic <- rep_len_default(italic, full_length, FALSE)
    weight <- rep_len_default(weight, full_length, "normal")
    width <- rep_len_default(width, full_length, "undefined")
    if (!all(c(length(path), length(index)) == 1)) {
      path <- rep_len(path, full_length)
      index <- rep_len(index, full_length)
    }
  }
  if (length(size) != 1) size <- rep_len_default(size, full_length, 12)
  if (length(res) != 1) res <- rep_len_default(res, full_length, 72)
  if (!all(file.exists(path))) {
    stop("path must point to a valid file", call. = FALSE)
  }
  variation <- add_standard_to_variations(
    rep_len(variation, full_length),
    italic = italic,
    weight = weight,
    width = width
  )
  get_font_info_c(
    path,
    as.integer(index),
    as.numeric(size),
    as.numeric(res),
    variation
  )
}
#' Query glyph-specific information from fonts
#'
#' This function allows you to extract information about the individual glyphs
#' in a font, based on a specified size. All size related measures are in
#' pixel-units. The function is vectorised to the length of the `glyphs` vector.
#'
#' @param glyphs A vector of glyphs. Strings will be split into separate glyphs
#' automatically
#' @inheritParams font_info
#' @param path,index path an index of a font file to circumvent lookup based on
#' family and style
#'
#' @return
#' A data.frame with information about each glyph, containing the following
#' columns:
#'
#' \describe{
#'   \item{glyph}{The glyph as a character}
#'   \item{index}{The index of the glyph in the font file}
#'   \item{width}{The width of the glyph}
#'   \item{height}{The height of the glyph}
#'   \item{x_bearing}{The horizontal distance from the origin to the leftmost part of the glyph}
#'   \item{y_bearing}{The vertical distance from the origin to the top part of the glyph}
#'   \item{x_advance}{The horizontal distance to move the cursor after adding the glyph}
#'   \item{y_advance}{The vertical distance to move the cursor after adding the glyph}
#'   \item{bbox}{The tight bounding box surrounding the glyph}
#' }
#'
#' @export
glyph_info <- function(
  glyphs,
  family = '',
  italic = FALSE,
  weight = "normal",
  width = "undefined",
  size = 12,
  res = 72,
  path = NULL,
  index = 0,
  variation = font_variation(),
  bold = deprecated()
) {
  if (is_font_variation(variation)) variation <- list(variation)
  n_strings <- length(glyphs)
  glyphs <- strsplit(glyphs, '')
  n_glyphs <- lengths(glyphs)
  glyphs <- unlist(glyphs)
  if (is.null(path)) {
    if (lifecycle::is_present(bold)) {
      lifecycle::deprecate_soft(
        "1.2.4",
        "glyph_info(bold)",
        "glyph_info(weight)"
      )
      weight <- ifelse(bold, "bold", "normal")
    }
    italic <- rep_len_default(italic, n_strings, FALSE)
    weight <- rep_len_default(weight, n_strings, "normal")
    width <- rep_len_default(width, n_strings, "undefined")
    fonts <- match_fonts(
      family = rep_len_default(family, n_strings, ''),
      italic = italic,
      weight = weight,
      width = width
    )
    path <- rep(fonts$path, n_glyphs)
    index <- rep(fonts$index, n_glyphs)
    italic <- rep(italic, n_glyphs)
    weight <- rep(weight, n_glyphs)
    width <- rep(width, n_glyphs)
  } else {
    if (!all(c(length(path), length(index)) == 1)) {
      path <- rep(rep_len(path, n_strings), n_glyphs)
      index <- rep(rep_len(index, n_strings), n_glyphs)
    }
    italic <- rep(rep_len_default(italic, n_strings, FALSE), n_glyphs)
    weight <- rep(rep_len_default(weight, n_strings, "normal"), n_glyphs)
    width <- rep(rep_len_default(width, n_strings, "undefined"), n_glyphs)
  }
  if (length(size) != 1) size <- rep(rep_len_default(size, n_strings, 12), n_glyphs)
  if (length(res) != 1) res <- rep(rep_len_default(res, n_strings, 72), n_glyphs)
  if (!all(file.exists(path))) {
    stop("path must point to a valid file", call. = FALSE)
  }
  variation <- add_standard_to_variations(
    rep(rep_len(variation, n_strings), n_glyphs),
    italic = italic,
    weight = weight,
    width = width
  )
  get_glyph_info_c(
    glyphs,
    path,
    as.integer(index),
    as.numeric(size),
    as.numeric(res),
    variation
  )
}

# Generated by cpp11: do not edit by hand

dev_string_widths_c <- function(string, family, face, size, cex, unit) {
  .Call(`_systemfonts_dev_string_widths_c`, string, family, face, size, cex, unit)
}

dev_string_metrics_c <- function(string, family, face, size, cex, unit) {
  .Call(`_systemfonts_dev_string_metrics_c`, string, family, face, size, cex, unit)
}

load_emoji_codes_c <- function(all, default_text, base_mod) {
  invisible(.Call(`_systemfonts_load_emoji_codes_c`, all, default_text, base_mod))
}

emoji_split_c <- function(string, path, index) {
  .Call(`_systemfonts_emoji_split_c`, string, path, index)
}

get_fallback_c <- function(path, index, string, variations) {
  .Call(`_systemfonts_get_fallback_c`, path, index, string, variations)
}

add_local_fonts <- function(paths) {
  .Call(`_systemfonts_add_local_fonts`, paths)
}

clear_local_fonts_c <- function() {
  invisible(.Call(`_systemfonts_clear_local_fonts_c`))
}

match_font_c <- function(family, italic, bold) {
  .Call(`_systemfonts_match_font_c`, family, italic, bold)
}

locate_fonts_c <- function(family, italic, weight, width) {
  .Call(`_systemfonts_locate_fonts_c`, family, italic, weight, width)
}

system_fonts_c <- function() {
  .Call(`_systemfonts_system_fonts_c`)
}

reset_font_cache_c <- function() {
  invisible(.Call(`_systemfonts_reset_font_cache_c`))
}

get_font_info_c <- function(path, index, size, res, variations) {
  .Call(`_systemfonts_get_font_info_c`, path, index, size, res, variations)
}

get_glyph_info_c <- function(glyphs, path, index, size, res, variations) {
  .Call(`_systemfonts_get_glyph_info_c`, glyphs, path, index, size, res, variations)
}

get_glyph_outlines <- function(glyph, path, index, size, variations, tolerance, verbose) {
  .Call(`_systemfonts_get_glyph_outlines`, glyph, path, index, size, variations, tolerance, verbose)
}

get_glyph_bitmap <- function(glyph, path, index, size, res, variations, color, verbose) {
  .Call(`_systemfonts_get_glyph_bitmap`, glyph, path, index, size, res, variations, color, verbose)
}

register_font_c <- function(family, paths, indices, features, settings) {
  invisible(.Call(`_systemfonts_register_font_c`, family, paths, indices, features, settings))
}

clear_registry_c <- function() {
  invisible(.Call(`_systemfonts_clear_registry_c`))
}

registry_fonts_c <- function() {
  .Call(`_systemfonts_registry_fonts_c`)
}

axes_to_tags <- function(axes) {
  .Call(`_systemfonts_axes_to_tags`, axes)
}

tags_to_axes <- function(tags) {
  .Call(`_systemfonts_tags_to_axes`, tags)
}

values_to_fixed <- function(values) {
  .Call(`_systemfonts_values_to_fixed`, values)
}

fixed_to_values <- function(fixed) {
  .Call(`_systemfonts_fixed_to_values`, fixed)
}

get_string_shape_c <- function(string, id, path, index, size, res, lineheight, align, hjust, vjust, width, tracking, indent, hanging, space_before, space_after) {
  .Call(`_systemfonts_get_string_shape_c`, string, id, path, index, size, res, lineheight, align, hjust, vjust, width, tracking, indent, hanging, space_before, space_after)
}

get_line_width_c <- function(string, path, index, size, res, include_bearing) {
  .Call(`_systemfonts_get_line_width_c`, string, path, index, size, res, include_bearing)
}

---
title: "Using systemfonts to handle fonts in R"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using systemfonts to handle fonts in R}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

<style type='text/css'>
pre {
  text-wrap: nowrap;
  overflow-x: scroll;
}
.figcaption {
  text-align: center;
  margin-bottom: 1em;
}
.pagedtable-wrapper, .fig {
  max-width: 99%;
  border: none;
  margin: 1em auto;
}
</style>

```{r, include = FALSE}
google <- suppressWarnings(try(readLines("https://8.8.8.8", n = 1L), silent = TRUE))
is_online <- !inherits(google, "try-error")

knitr::opts_chunk$set(
  dev = "ragg_png",
  dpi = 144,
  collapse = TRUE,
  comment = "#>",
  fig.asp = NULL,
  fig.height = 4.326,
  fig.width = 7,
  eval = is_online
)

systemfonts::require_font("Spectral", fallback = "serif")
```

```{asis, echo = !is_online}
> This vignette was compiled with no internet access. Since many of the code chunks require access to online font repositories evaluation has been turned off.
```

> This text expects a basic understanding of fonts and typography. If you feel like you could use a brush-up you can consult [this light introduction to the subject](fonts_basics.html).

systemfonts is designed to give R a modern text rendering stack. That's unfortunately impossible without coordination with the graphics device, which means that to use all these features you need a supported graphics device. There are currently two options:

-   The [ragg](https://ragg.r-lib.org) package provides graphics devices for rendering raster graphics in a variety of formats (PNG, JPEG, TIFF) and uses systemfonts and textshaping extensively.
-   The [svglite](https://svglite.r-lib.org) package provides a graphic device for rendering vector graphics to SVG using systemfonts and textshaping for text.

You might notice there's currently a big hole in this workflow: PDFs. This is something we plan to work on in the future.

## A systemfonts based workflow

With all that said, how do you actually use systemfonts to use custom fonts in your plots? First, you'll need to use ragg or svglite.

### Using ragg

While there is no way to unilaterally make `ragg::agg_png()` the default everywhere, it's possible to get close:

-   Positron: recent versions automatically use ragg for the plot pane if it's installed.

-   RStudio IDE: set "AGG" as the backend under Global Options \> General \> Graphics.

-   `ggplot2::ggsave()`: ragg will be automatically used for raster output if installed.

-   R Markdown and Quarto: you need to set the `dev` option to `"ragg_png"`. You can either do this with code:

    ``` r
    #| include: false
    knitr::opts_chunk$set(dev = "ragg_png")
    ```

    Or in Quarto, you can set it in the yaml metadata:

    ``` yaml
    ---
    title: "My Document"
    format: html
    knitr:
      opts_chunk:
        dev: "ragg_png"
    ---
    ```

If you want to use a font installed on your computer, you're done!

```{r}
#| fig.height: 1
#| out.extra: class='fig'
grid::grid.text(
  "Spectral 🎉",
  gp = grid::gpar(fontfamily = "Spectral", fontface = 2, fontsize = 30)
)
```

Or, if using ggplot2

```{r}
#| out.extra: class='fig'
#| eval: !expr exists("penguins")
library(ggplot2)
ggplot(na.omit(penguins)) +
  geom_point(aes(x = bill_len, y = body_mass, colour = species)) +
  labs(x = "Bill Length", y = "Body Mass", colour = "Species") +
  theme_minimal(base_family = "Spectral")
```

If the results don't look as you expect, you can use various systemfonts helpers to diagnose the problem:

```{r}
systemfonts::match_fonts("Spectral", weight = "bold")
systemfonts::font_fallback("🎉", family = "Spectral", weight = "bold")
```

If you want to see all the fonts that are available for use, you can use `systemfonts::system_fonts()`

```{r}
#| eval: false
systemfonts::system_fonts()
```

```{r}
#| echo: false
all_fonts <- systemfonts::system_fonts()
all_fonts <- all_fonts[!grepl("^/Users", all_fonts$path),]
rmarkdown::paged_table(all_fonts)
```

### Extra font styles

As we discussed above, the R interface only allows you to select between four styles: plain, italic, bold, and bold-italic. If you want to use a thin font, you have no way of communicating this wish to the device. To overcome this, systemfonts provides `register_variant()` which allows you to register a font with a new typeface name. For example, to use the light font from the Spectral typeface you can register it as follows:

```{r}
systemfonts::register_variant(
  name = "Spectral Light",
  family = "Spectral",
  weight = "light"
)
```

Now you can use Spectral Light where you would otherwise specify the typeface:

```{r}
#| fig.height: 1
#| out.extra: class='fig'
grid::grid.text(
  "Light weight is soo classy",
  gp = grid::gpar(fontfamily = "Spectral Light", fontsize = 30)
)
```

`register_variant()` also allows you to turn on font features otherwise hidden away:

```{r}
#| fig.height: 1
#| out.extra: class='fig'
systemfonts::register_variant(
  name = "Spectral Small Caps",
  family = "Spectral",
  features = systemfonts::font_feature(
    letters = "small_caps"
  )
)
grid::grid.text(
  "All caps — Small caps",
  gp = grid::gpar(fontfamily = "Spectral Small Caps", fontsize = 30)
)
```

### Fonts from other places

Historically, systemfonts primary role was to access the font installed on your computer, the **system fonts**. But what if you're using a computer where you don't have the rights to install new fonts, or you don't want the hassle of installing a font just to use it for a single plot? That's the problem solved by `systemfonts::add_font()` which makes it easy to use a font based on a path. But in many cases you don't even need that as systemfont now scans `./fonts` and `~/fonts` and adds any font files it find. This means that you can put personal fonts in a fonts folder in your home directory, and project fonts in a fonts directory at the root of the project. This is a great way to ensure that specific fonts are available when you deploy some code to a server.

And you don't even need to leave R to populate these folders. `systemfonts::get_from_google_fonts()` will download and install a google font in `~/fonts`:

```{r}
#| eval: false
systemfonts::get_from_google_fonts("Barrio")

grid::grid.text(
  "A new font a day keeps Tufte away",
  gp = grid::gpar(fontfamily = "Barrio", fontsize = 30)
)
```

```{r}
#| echo: false
#| message: false
#| fig.height: 1
#| out.extra: class='fig'
systemfonts::require_font("Barrio")
grid::grid.text(
  "A new font a day keeps Tufte away",
  gp = grid::gpar(fontfamily = "Barrio", fontsize = 30)
)
```

And if you want to make sure this code works for anyone using your code (regardless of whether or not they already have the font installed), you can use `systemfonts::require_font()`. If the font isn't already installed, this function download it from one of the repositories it knows about. If it can't find it it will either throw an error (the default) or remap the name to another font so that plotting will still succeed.

```{r}
#| fig.height: 1.5
#| out.extra: class='fig'
systemfonts::require_font("Rubik Distressed")

grid::grid.text(
  "There are no bad fonts\nonly bad text",
  gp = grid::gpar(fontfamily = "Rubik Distressed", fontsize = 30)
)
```

By default, `require_font()` places new fonts in a temporary folder so it doesn't pollute your carefully curated collection of fonts.

### Font embedding in SVG

Fonts work a little differently in vector formats like SVG. These formats include the raw text and only render the font when you open the file. This makes for small, accessible files with crisp text at every level of zoom. But it comes with a price: since the text is rendered when it's opened, it relies on the font in use being available on the viewer's computer. This obviously puts you at the mercy of their font selection, so if you want consistent outputs you'll need to **embed** the font.

In SVG, you can embed fonts using an `@import` statement in the stylesheet, and can point to a web resource so the SVG doesn't need to contain the entire font. systemfonts provides facilities to generate URLs for import statements and can provide them in a variety of formats:

```{r}
systemfonts::fonts_as_import("Barrio")
systemfonts::fonts_as_import("Rubik Distressed", type = "link")
```

Further, if the font is not available from a given online repository, it can embed the font data directly into the URL:

```{r}
substr(systemfonts::fonts_as_import("Arial", repositories = NULL), 1, 200)
```

svglite uses this feature to allow seamless font embedding with the `web_fonts` argument. It can take a URL as returned by `fonts_as_import()` or just the name of the typeface and the URL will automatically be resolved. Look at line 6 in the SVG generated below

```{r}
svg <- svglite::svgstring(web_fonts = "Barrio")
grid::grid.text("Example", gp = grid::gpar(fontfamily = "Barrio"))
invisible(dev.off())
svg()
```

## Want more?

This text has mainly focused on how to use the fonts you desire from within R. R has other limitations when it comes to text rendering specifically how to render text that consists of a mix of fonts. This has been solved by [marquee](https://marquee.r-lib.org) and the curious soul can continue there in order to up their skills in rendering text with R.

---
title: "systemfonts C interface"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{systemfonts C interface}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r}
#| label: setup
library(systemfonts)
```

Most of the functionality in systemfonts is intended to be used from compiled
code to help e.g. graphic devices to resolve font specifications to a font file
prior to rendering. systemfonts provide key functionality to get called at the
C level by putting systemfonts in the `LinkingTo` field in the description and
adding `#include <systemfonts.h>` to your C code. Make sure systemfonts is
loaded before using it, e.g. by having `match_fonts()` imported into your package
namespace. All functions are provided in the `systemfonts::ver2` namespace.
Legacy API is not namespaced. The different functionality will be discussed
below:

## Font matching
The C equivalent of the `match_fonts()` R function is `locate_font()` with the
following signature:

```C
FontSettings2 locate_font(
  const char *family,
  double italic,
  double weight,
  double width,
  const int* axes,
  const int* coords,
  int n_axes
)
```

It takes a UTF-8 encoded string with the font family name, a double giving
italic (usually 0.0 == "upright" and 1.0 == "italic"), a double giving weight
(usually ranging between 100.0 and 1000.0 — 0.0 means "undefined") and a double
giving "width" (usually ranging from 1.0 to 10.0 — 0.0 means undefined). Lastly
you can provide variable axis coords with the `axes` and `coords` array pointers
with `n_axes` giving the number in the arrays (which are assumed to be of the
same length). The values of each array are not immediately understandable to the
human eye and will usually come from a user through a call to
`font_variation()`. If the axes array contain "ital", "wght", and/or "wdth"
*and* the font has these variable axes then the values for these axes will
overwrite the values provide in `italic`, `weight`, and `width`.

The returned `FontSettings2` struct will contain both the font location and
index along with any OpenType feature settings and the axes settings in the case
of a variable font. The struct (along with its `FontFeature` struct dependency)
is shown below and is pretty self-documenting.

Do not cache the `FontSettings2` struct as the `features`, `axes`, and `coords`
arrays  may be cleared at any time after the call has ended. systemfonts itself
takes care of caching so this is not something you should be concerned with in
your code.

```C
struct FontFeature {
  char feature[4];
  int setting;
};
struct FontSettings {
  char file[PATH_MAX + 1];
  unsigned int index;
  const FontFeature* features;
  int n_features;
  const int* axes;
  const int* coords;
  int n_axes;
};
```

## Glyph metrics
The C equivalent of `glyph_info()` is `glyph_metrics()` with the following
signature:

```C
int glyph_metrics(
  uint32_t code,
  const FontSettings2& font,
  double size,
  double res,
  double* ascent,
  double* descent,
  double* width
)
```

It takes the glyph to measure as an int giving the UTF code of the glyph, with a
`FontSettings2` object describing the font. Further it takes a size in pt and a
resolution in ppi. It will write the ascent, descent, and width in pts to the
pointers passed in, and return `0` if the operation was successful.

## Retrieving cached freetype face
A heavy part of text layouting is reading and parsing font files. systemfonts
contains its own cache to make sure that parsing is kept at a minimum. If you
want to use this cache to load and cache freetype face object (FT_Face) you can
use `get_cached_face()`. This resides in a separate header (`systemfonts-ft.h`)
because it requires FreeType to be linked in your package, which the rest of the
C api does not. It will look in the cache for a face and size that
matches your request and return that if found. If not, it will load it for you
and add it to the cache, before returning it to you. `get_cached_face()` sets
the passed int error pointer to 0 if successful.

```C
get_cached_face(
  const FontSettings2& font,
  double size,
  double res,
  int * error
)
```

Freetype uses reference counting to keep track of objects and the count is
increased by a call to `get_cached_face()`. It is the responsibility of the
caller to decrease it once the face is no longer needed using `FT_Done_Face()`.

## Check for Freetype compatibility
If you are using a cached face from systemfonts you should ensure that your code
has been compiled with the same version of Freetype as systemfonts has. You can
do this with the `check_ft_version()` from the `systemfonts-ft.h` header. It takes no arguments and return `true` if the Freetype version from systemfonts 
corresponds with the one your library is compiled with.

## Font fallback
When rendering text it is not given that all the requested characters have a
glyph in the given font. While one can elect to render a "missing glyph" glyph
(often either an empty square or a questionmark in a tilted square) a better
approach is often to find a font substitute that does contain the character and
use that for rendering it. This function allows you to find a fallback font for
a given string and font. The string should be stripped of characters that you
already know how to render. The fallback font is returned as a `FontSettings2`
object, though features are always empty.

```C
FontSettings2 get_fallback(
  const char* string,
  const FontSettings2& font
)
```

## Font Weight
When encoding text with CSS it may be necessary to know the exact weight of the
font given by a file so that it may be reflected in the style sheet. This
function takes a `FontSettings2` object and returns the weight (100-900 or 0 if
it is undefined by the font) respecting the variable axes settings if given.

```C
int get_font_weight(
  const FontSettings2& font
)
```

## Family name
It may be beneficial to know the family name from a given font. This can be
obtained with `get_font_family()` which will write the name to the provided
`char*` argument. It will return 0 if it was somehow unsuccessful.

```C
int get_font_family(
  const FontSettings2& font,
  char* family,
  int max_length
)
```

## Emoji location
Figuring out which character in a string should be treated as an emoji is
non-trivial due to the existence of emojis with text representation default etc.
systemfonts allow you to get the embedding of emojis in a string based on the
correct rules.

```C
void detect_emoji_embedding(
  const uint32_t* string,
  int n,
  int* embedding,
  const FontSettings2& font
)
```

---
title: "systemfonts C interface"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{systemfonts C interface}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(systemfonts)
```

Most of the functionality in systemfonts is intended to be used from compiled
code to help e.g. graphic devices to resolve font specifications to a font file
prior to rendering. systemfonts provide key functionality to get called at the 
C level by putting systemfonts in the `LinkingTo` field in the description and
adding `#include <systemfonts.h>` to your C code. Make sure systemfonts is 
loaded before using it, e.g. by having `match_font()` imported into your package
namespace. The different functionality will be discussed below

## Font matching
The C equivalent of the `match_font()` R function is `locate_font()` with the
following signature:

```C
int locate_font(
  const char *family, 
  int italic, 
  int bold, 
  char *path, 
  int max_path_length
)
```

It takes a UTF-8 encoded string with the font family name, an int setting both
italic and bold styles along with a char pointer to be filled with the located
path and the maximum length it can hold. The return value is an int giving the
index of the font in the font file.

With the advent of systemfonts 0.3.0 fonts can now also have OpenType features
attached to them through the use of `register_font()` or `register_variant()`.
If you wish to support such features you can use an alternative to the above:

```C
FontSettings locate_font_with_features(
  const char *family, 
  int italic, 
  int bold
)
```

The returned `FontSettings` struct will contain both the font location and index
along with any OpenType feature settings. The struct (along with its 
`FontFeature` struct dependency) is shown below and is pretty self-documenting.

Do not cache the `FontSettings` struct as the `features` array may be cleared at
any time after the call has ended. systemfonts itself takes care of caching so
this is not something you should be concerned with in your code.

```C
struct FontFeature {
  char feature[4];
  int setting;
};
struct FontSettings {
  char file[PATH_MAX + 1];
  unsigned int index;
  const FontFeature* features;
  int n_features;
};
```

## Glyph metrics
The C equivalent of `glyph_info()` is `glyph_metrics()` with the following 
signature:

```C
int glyph_metrics(
  uint32_t code, 
  const char* fontfile, 
  int index, 
  double size, 
  double res, 
  double* ascent, 
  double* descent, 
  double* width
)
```

It takes the glyph to measure as an int giving the UTF code of the glyph, along
with a fontfile and index to identify the font to measure with. Further it takes
a size in pt and a resolution in ppi. It will write the ascent, descent, and 
width in pts to the pointers passed in, and return `0` if the operation was 
successful.

## String width
The C equivalent of the `string_width()` R function is also called 
`string_width()` with the following signature:

```C
string_width(
  const char* string, 
  const char* fontfile, 
  int index, 
  double size, 
  double res, 
  int include_bearing, 
  double* width
)
```

This function calculates the width of a string, ignoring any newlines (these are
automatically being handled by the graphic engine). It takes a UTF-8 encoded 
string, along with a fontfile and index identifying the font to use for the 
calculation. It also take a size in pt and a res in ppi for setting the size. In
addition it takes an include_bearing flag to control whether the bearings of the
first and last character should be taken into account (this is recommended by 
the graphic engine). It will write the width in pts to the passed in pointer and 
return 0 if successful.

## String shape
A parred down version of `shape_string()` is accessible at the C level with 
`string_shape()`. It behaves more or less like `string_width()` above, but 
instead returns the location to write each glyph at relative to a (0, 0) origin.

```C
string_shape(
  const char* string, 
  const char* fontfile, 
  int index, 
  double size, 
  double res, 
  double* x, 
  double* y, 
  unsigned int max_length
)
```

`string_shape()` behaves more or less like `string_width()` above, but instead 
returns the location to write each glyph at relative to a (0, 0) origin. It 
takes a UTF-8 encoded string, along with a fontfile and index identifying the 
font to use for the calculation. It also take a size in pt and a res in ppi for 
setting the size. In addition it takes an include_bearing flag to control 
whether the bearings of the first and last character should be taken into 
account (this is recommended by the graphic engine). It will write the x and y
location of each glyph in pts to the passed in arrays, stopping before the 
provided max_length and return 0 if successful.

## Retrieving cached freetype face
A heavy part of text layouting is reading and parsing font files. systemfonts 
contains its own cache to make sure that parsing is kept at a minimum. If you
want to use this cache to load and cache freetype face object (FT_Face) you can
use `get_cached_face()`. This resides in a separate header (`systemfonts-ft.h`) 
because it requires FreeType to be linked in your package, which the rest of the
C api does not. It will look in the cache for a face and size that 
matches your request and return that if found. If not, it will load it for you 
and add it to the cache, before returning it to you. `get_cached_face()` sets 
the passed in error pointer to 0 if successful.

```C
get_cached_face(
  const char* fontfile, 
  int index, 
  double size, 
  double res,
  int * error
)
```

Freetype uses reference counting to keep track of objects and the count is 
increased by a call to `get_cached_face()`. It is the responsiblity of the 
caller to decrease it once the face is no longer needed using `FT_Done_Face()`.

## Font fallback
When rendering text it is not given that all the requested characters have a 
glyph in the given font. While one can elect to render a "missing glyph" glyph 
(often either an empty square or a questionmark in a tilted square) a better 
approach is often to find a font substitute that does contain the character and
use that for rendering it. This function allows you to find a fallback font for 
a given string and font. The string should be stripped of characters that you 
already know how to render. The fallback font is returned as a FontSettings 
object, though features are always empty.

```C
FontSettings get_fallback(
  const char* string,
  const char* path,
  int index
)
```

## Font Weight
When encoding text with CSS it may be necessary to know the exact weight of the
font given by a file so that it may be reflected in the style sheet. This 
function takes a path and an index and returns the weight (100-900 in steps of
100) or 0 if it is undefined by the font.

```C
int get_font_weight(
  const char* path,
  int index
)
```

## Family name
It may be beneficial to know the family name from a given path and index into a
font. This can be obtained with `get_font_family()` which will write the name to
the provided `char*` argument. It will return 0 if it was somehow unsuccessful.

```C
int get_font_family(
  const char* path,
  int index,
  char* family,
  int max_length
)
```

## Emoji location
Figuring out which character in a string should be treated as an emoji is 
non-trivial due to the existence of emojis with text representation default etc.
systemfonts allow you to get the embedding of emojis in a string based on the 
correct rules.

```C
void detect_emoji_embedding(
  const uint32_t* string, 
  int n, 
  int* embedding, 
  const char *path, 
  int index
)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/register_font.R
\name{register_variant}
\alias{register_variant}
\title{Register a font as a variant as an existing one}
\usage{
register_variant(
  name,
  family,
  weight = NULL,
  width = NULL,
  features = font_feature()
)
}
\arguments{
\item{name}{The new family name the variant should respond to}

\item{family}{The name of an existing font family that this is a variant of}

\item{weight}{One or two of \code{"thin"}, \code{"ultralight"}, \code{"light"}, \code{"normal"},
\code{"medium"}, \code{"semibold"}, \code{"bold"}, \code{"ultrabold"}, or \code{"heavy"}. If one is
given it sets the weight for the whole variant. If two is given the first
one defines the plain weight and the second the bold weight. If \code{NULL} then
the variants of the given family closest to \code{"normal"} and \code{"bold"} will be
chosen.}

\item{width}{One of \code{"ultracondensed"}, \code{"extracondensed"}, \code{"condensed"},
\code{"semicondensed"}, \code{"normal"}, \code{"semiexpanded"}, \code{"expanded"},
\code{"extraexpanded"}, or \code{"ultraexpanded"} giving the width of the variant. If
\code{NULL} then the width closest to \code{"normal"} will be chosen.}

\item{features}{A \code{\link{font_feature}} object describing the specific OpenType
font features to turn on for the registered font variant.}
}
\description{
This function is a wrapper around \code{\link[=register_font]{register_font()}} that allows you to easily
create variants of existing system fonts, e.g. to target different weights
and/or widths, or for attaching OpenType features to a font.
}
\section{Font matching}{
During font matching, systemfonts has to look in three different locations.
The font registry (populated by \code{\link[=register_font]{register_font()}}/\code{\link[=register_variant]{register_variant()}}), the
local fonts (populated with \code{\link[=add_fonts]{add_fonts()}}/\code{\link[=scan_local_fonts]{scan_local_fonts()}}), and the
fonts installed on the system. It does so in that order: registry > local >
installed.

The matching performed at each step also differs. The fonts in the registry
is only matched by family name. The local fonts are matched based on all the
provided parameters (family, weight, italic, etc) in a way that is local to
systemfonts, but try to emulate the system native matching. The installed
fonts are matched using the system native matching functionality on macOS and
Linux. On Windows the installed fonts are read from the system registry and
matched using the same approach as for local fonts. Matching will always find
a font no matter what you throw at it, defaulting to "sans" if nothing else
is found.
}

\examples{
# Get the default "sans" family
sans <- match_fonts("sans")$path
sans <- system_fonts()$family[system_fonts()$path == sans][1]

# Register a variant of it:
register_variant(
  "sans_ligature",
  sans,
  features = font_feature(ligatures = "discretionary")
)

registry_fonts()

# clean up
clear_registry()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_info.R
\name{font_info}
\alias{font_info}
\title{Query font-specific information}
\usage{
font_info(
  family = "",
  italic = FALSE,
  weight = "normal",
  width = "undefined",
  size = 12,
  res = 72,
  path = NULL,
  index = 0,
  variation = font_variation(),
  bold = deprecated()
)
}
\arguments{
\item{family}{The name of the font families to match}

\item{italic}{logical indicating the font slant}

\item{weight}{The weight to query for, either in numbers (\code{0}, \code{100}, \code{200},
\code{300}, \code{400}, \code{500}, \code{600}, \code{700}, \code{800}, or \code{900}) or strings (\code{"undefined"},
\code{"thin"}, \code{"ultralight"}, \code{"light"}, \code{"normal"}, \code{"medium"}, \code{"semibold"},
\code{"bold"}, \code{"ultrabold"}, or \code{"heavy"}). \code{NA} will be interpreted as
\code{"undefined"}/\code{0}}

\item{width}{The width to query for either in numbers (\code{0}, \code{1}, \code{2},
\code{3}, \code{4}, \code{5}, \code{6}, \code{7}, \code{8}, or \code{9}) or strings (\code{"undefined"},
\code{"ultracondensed"}, \code{"extracondensed"}, \code{"condensed"}, \code{"semicondensed"},
\code{"normal"}, \code{"semiexpanded"}, \code{"expanded"}, \code{"extraexpanded"}, or
\code{"ultraexpanded"}). \code{NA} will be interpreted as \code{"undefined"}/\code{0}}

\item{size}{The pointsize of the font to use for size related measures}

\item{res}{The ppi of the size related measures}

\item{path, index}{path and index of a font file to circumvent lookup based on
family and style}

\item{variation}{A \code{font_variation} object or a list of them to control
variable fonts}

\item{bold}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{weight = "bold"} instead}
}
\value{
A data.frame giving info on the requested font + size combinations. The
data.frame will contain the following columns:

\describe{
\item{path}{The path to the font file}
\item{index}{The 0-based index of the font in the fontfile}
\item{family}{The family name of the font}
\item{style}{The style name of the font}
\item{name}{The name of the font, if present, otherwise the family name}
\item{italic}{A logical giving if the font is italic}
\item{bold}{A logical giving if the font is bold}
\item{monospace}{A logical giving if the font is monospace}
\item{weight}{A factor giving the weight of the font}
\item{width}{A factor giving the width of the font}
\item{kerning}{A logical giving if the font supports kerning}
\item{color}{A logical giving if the font has color glyphs}
\item{scalable}{A logical giving if the font is scalable}
\item{vertical}{A logical giving if the font is vertical}
\item{n_glyphs}{The number of glyphs in the font}
\item{n_sizes}{The number of predefined sizes in the font}
\item{n_charmaps}{The number of character mappings in the font file}
\item{bbox}{A bounding box large enough to contain any of the glyphs in the font}
\item{max_ascend}{The maximum ascend of the tallest glyph in the font}
\item{max_descent}{The maximum descend of the most descending glyph in the font}
\item{max_advance_width}{The maximum horizontal advance a glyph can make}
\item{max_advance_height}{The maximum vertical advance a glyph can make}
\item{lineheight}{The height of a single line of text in the font}
\item{underline_pos}{The position of a potential underlining segment}
\item{underline_size}{The width the the underline}
}
}
\description{
Get general information about a font, relative to a given size. Size specific
measures will be returned in pixel units. The function is vectorised to the
length of the longest argument.
}
\examples{
font_info('serif')

# Avoid lookup if font file is already known
sans <- match_fonts('sans')
font_info(path = sans$path, index = sans$index)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{get_fallback}
\alias{get_fallback}
\title{Get location of the fallback font}
\usage{
get_fallback()
}
\description{
Get location of the fallback font
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_feature.R
\name{font_feature}
\alias{font_feature}
\title{Define OpenType font feature settings}
\usage{
font_feature(ligatures = NULL, letters = NULL, numbers = NULL, ...)
}
\arguments{
\item{ligatures}{Settings related to ligatures. One or more types of
ligatures to turn on (see details).}

\item{letters}{Settings related to the appearance of single
letters (as opposed to ligatures that substitutes multiple letters). See
details for supported values.}

\item{numbers}{Settings related to the appearance of numbers. See details for
supported values.}

\item{...}{key-value pairs with the key being the 4-letter tag and the value
being the setting (usually \code{TRUE} to turn it on).}
}
\value{
A \code{font_feature} object
}
\description{
This function encapsulates the specification of OpenType font features. Some
specific features have named arguments, but all available features can be
set by using its specific 4-letter tag For a list of the 4-letter tags
available see e.g. the overview on
\href{https://en.wikipedia.org/wiki/List_of_typographic_features}{Wikipedia}.
}
\details{
OpenType features are defined by a 4-letter tag along with an integer value.
Often that value is a simple \code{0} (off) or \code{1} (on), but some features support
additional values, e.g. stylistic alternates (\code{salt}) where a font may
provide multiple variants of a letter and the value will be used to chose
which one to use.

Common features related to appearance may be given with a long form name to
either the \code{ligatures}, \code{letters}, or \code{numbers} argument to avoid remembering
the often arbitrary 4-letter tag. Providing a long form name is the same as
setting the tag to \code{1} and can thus not be used to set tags to other values.

The possible long form names are given below with the tag in parenthesis:

\strong{Ligatures}
\itemize{
\item \code{standard} (\emph{liga}): Turns on standard multiple letter substitution
\item \code{historical} (\emph{hlig}): Use obsolete historical ligatures
\item \code{contextual} (\emph{clig}): Apply secondary ligatures based on the character
patterns surrounding the potential ligature
\item \code{discretionary} (\emph{dlig}): Use ornamental ligatures
}

\strong{Letters}
\itemize{
\item \code{swash} (\emph{cswh}): Use contextual swashes (ornamental decorations)
\item \code{alternates} (\emph{calt}): Use alternate letter forms based on the surrounding
pattern
\item \code{historical} (\emph{hist}): Use obsolete historical forms of the letters
\item \code{localized} (\emph{locl}): Use alternate forms preferred by the script language
\item \code{randomize} (\emph{rand}): Use random variants of the letters (e.g. to mimic
handwriting)
\item \code{alt_annotation} (\emph{nalt}): Use alternate annotations (e.g. circled digits)
\item \code{stylistic} (\emph{salt}): Use a stylistic alternative form of the letter
\item \code{subscript} (\emph{subs}): Set letter in subscript
\item \code{superscript} (\emph{sups}): Set letter in superscript
\item \code{titling} (\emph{titl}): Use letter forms well suited for large text and titles
\item \code{small_caps} (\emph{smcp}): Use small caps variants of the letters
}

\strong{Numbers}
\itemize{
\item \code{lining} (\emph{lnum}): Use number variants that rest on the baseline
\item \code{oldstyle} (\emph{onum}): Use old style numbers that use descender and ascender
for various numbers
\item \code{proportional} (\emph{pnum}): Let numbers take up width based on the visual
width of the glyph
\item \code{tabular} (\emph{tnum}): Enforce all numbers to take up the same width
\item \code{fractions} (\emph{frac}): Convert numbers separated by \code{/} into a fraction
glyph
\item \code{fractions_alt} (\emph{afrc}): Use alternate fraction form with a horizontal
divider
}
}
\examples{
font_feature(letters = "stylistic", numbers = c("lining", "tabular"))

# Use the tag directly to access additional stylistic variants
font_feature(numbers = c("lining", "tabular"), salt = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/emoji.R
\name{str_split_emoji}
\alias{str_split_emoji}
\title{Split a string into emoji and non-emoji glyph runs}
\usage{
str_split_emoji(
  string,
  family = "",
  italic = FALSE,
  bold = FALSE,
  path = NULL,
  index = 0
)
}
\arguments{
\item{string}{A character vector of strings that should be splitted.}

\item{family}{The name of the font family}

\item{italic}{logicals indicating the font style}

\item{bold}{logicals indicating the font style}

\item{path, index}{path an index of a font file to circumvent lookup based on
family and style}
}
\value{
A data.frame containing the following columns:
\describe{
\item{string}{The substring containing a consecutive run of glyphs}
\item{id}{The index into the original \code{string} vector that the substring is part of}
\item{emoji}{A logical vector giving if the substring is a run of emojis or not}
}
}
\description{
In order to do correct text rendering, the font needed must be figured out. A
common case is rendering of emojis within a string where the system emoji
font is used rather than the requested font. This function will inspect the
provided strings and split them up in runs that must be rendered with the
emoji font, and the rest. Arguments are recycled to the length of the \code{string}
vector.
}
\examples{
emoji_string <- "This is a joke\U0001f642. It should be obvious from the smiley"
str_split_emoji(emoji_string)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/web_fonts.R
\name{search_web_fonts}
\alias{search_web_fonts}
\title{Search font repositories for a font based on family name}
\usage{
search_web_fonts(family, n_max = 10, ...)
}
\arguments{
\item{family}{The font family name to look for}

\item{n_max}{The maximum number of matches to return}

\item{...}{
  Arguments passed on to \code{\link[utils:adist]{utils::adist}}
  \describe{
    \item{\code{costs}}{a numeric vector or list with names partially matching
    \samp{insertions}, \samp{deletions} and \samp{substitutions} giving
    the respective costs for computing the Levenshtein distance, or
    \code{NULL} (default) indicating using unit cost for all three
    possible transformations.}
    \item{\code{counts}}{a logical indicating whether to optionally return the
    transformation counts (numbers of insertions, deletions and
    substitutions) as the \code{"counts"} attribute of the return
    value.}
    \item{\code{fixed}}{a logical.  If \code{TRUE} (default), the \code{x}
    elements are used as string literals.  Otherwise, they are taken as
    regular expressions and \code{partial = TRUE} is implied
    (corresponding to the approximate string distance used by
    \code{\link{agrep}} with \code{fixed = FALSE}).}
    \item{\code{partial}}{a logical indicating whether the transformed \code{x}
    elements must exactly match the complete \code{y} elements, or only
    substrings of these.  The latter corresponds to the approximate
    string distance used by \code{\link{agrep}} (by default).}
    \item{\code{ignore.case}}{a logical.  If \code{TRUE}, case is ignored for
    computing the distances.}
    \item{\code{useBytes}}{a logical.  If \code{TRUE} distance computations are
    done byte-by-byte rather than character-by-character.}
  }}
}
\value{
A data.frame with the columns \code{family}, giving the family name of the
matched font, and \code{repository} giving the repository it was found in.
}
\description{
While it is often advisable to visit the webpage for a font repository when
looking for a font, in order to see examples etc, \code{search_web_fonts()}
provide a quick lookup based on family name in the repositories supported by
systemfonts (currently \href{https://fonts.google.com}{Google Fonts} and
\href{https://www.fontsquirrel.com}{Font Squirrel} - \href{https://fonts.bunny.net/}{Bunny Fonts}
provide the same fonts as Google Fonts but doesn't have a search API). The
lookup is based on fuzzy matching provided by \code{\link[utils:adist]{utils::adist()}} and the
matching parameters can be controlled through \code{...}
}
\examples{
# Requires an internet connection

# search_web_fonts("Spectral")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system_fonts.R
\name{system_fonts}
\alias{system_fonts}
\title{List all fonts installed on your system}
\usage{
system_fonts()
}
\value{
A data frame with a row for each font and various information in each
column
}
\description{
List all fonts installed on your system
}
\examples{
# See all monospace fonts
fonts <- system_fonts()
fonts[fonts$monospace, ]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/systemfonts-package.R
\docType{package}
\name{systemfonts-package}
\alias{systemfonts}
\alias{systemfonts-package}
\title{systemfonts: System Native Font Finding}
\description{
Provides system native access to the font catalogue. As font
    handling varies between systems it is difficult to correctly locate 
    installed fonts across different operating systems. The 'systemfonts' 
    package provides bindings to the native libraries on Windows, macOS and 
    Linux for finding font files that can then be used further by e.g. graphic
    devices. The main use is intended to be from compiled code but 'systemfonts'
    also provides access from R.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/r-lib/systemfonts}
  \item Report bugs at \url{https://github.com/r-lib/systemfonts/issues}
}

}
\author{
\strong{Maintainer}: Thomas Lin Pedersen \email{thomas.pedersen@rstudio.com} (\href{https://orcid.org/0000-0002-5147-4711}{ORCID})

Authors:
\itemize{
  \item Jeroen Ooms \email{jeroen@berkeley.edu} (\href{https://orcid.org/0000-0002-4035-0289}{ORCID})
  \item Devon Govett (Author of font-manager)
}

Other contributors:
\itemize{
  \item RStudio [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/web_fonts.R
\name{require_font}
\alias{require_font}
\title{Ensure font availability in a script}
\usage{
require_font(
  family,
  fallback = NULL,
  dir = tempdir(),
  repositories = c("Google Fonts", "Font Squirrel", "Font Library"),
  error = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{family}{The font family to require}

\item{fallback}{An available font to fall back to if \code{family} cannot be found
or downloaded}

\item{dir}{The location to put the font file downloaded from repositories}

\item{repositories}{The repositories to search for the font in case it is not
available on the system. They will be tried in the order given. Currently
\code{"Google Fonts"}, \code{"Font Squirrel"}, and \code{"Font Library"} is available.}

\item{error}{Should the function throw an error if unsuccessful?}

\item{verbose}{Should status messages be emitted?}
}
\value{
Invisibly \code{TRUE} if the font is available or \code{FALSE} if not (this can
only be returned if \code{error = FALSE})
}
\description{
When running a script on a different machine you are not always in control of
which fonts are installed on the system and thus how graphics created by the
script ends up looking. \code{require_font()} is a way to specify your font
requirements for a script. It will look at the available fonts and if the
required font family is not present it will attempt to fetch it from one of
the given repositories (in the order given). If that fails, it will either
throw an error or, if \code{fallback} is given, provide an alias for the fallback
so it maps to the required font.
}
\examples{
# Should always work
require_font("sans")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/web_fonts.R
\name{web-fonts}
\alias{web-fonts}
\alias{get_from_google_fonts}
\alias{get_from_font_squirrel}
\title{Download and add web font}
\usage{
get_from_google_fonts(family, dir = "~/fonts", woff2 = FALSE)

get_from_font_squirrel(family, dir = "~/fonts")
}
\arguments{
\item{family}{The font family to download (case insensitive)}

\item{dir}{Where to download the font to. The default places it in your user
local font folder so that the font will be available automatically in new R
sessions. Set to \code{tempdir()} to only keep the font for the session.}

\item{woff2}{Should the font be downloaded in the woff2 format (smaller and
more optimized)? Defaults to FALSE as the format is not supported on all
systems}
}
\value{
A logical invisibly indicating whether a font was found and
downloaded or not
}
\description{
In order to use a font in R it must first be made available locally. These
functions facilitate the download and registration of fonts from online
repositories.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_outline.R
\name{glyph_raster_grob}
\alias{glyph_raster_grob}
\title{Convert an extracted glyph raster to a grob}
\usage{
glyph_raster_grob(glyph, x, y, ..., default.units = "bigpts")
}
\arguments{
\item{glyph}{The nativeRaster object returned as one of the elements by
\code{\link[=glyph_raster]{glyph_raster()}}}

\item{x, y}{The baseline location of the glyph}

\item{...}{
  Arguments passed on to \code{\link[grid:grid.raster]{grid::rasterGrob}}
  \describe{
    \item{\code{image}}{
    Any R object that can be coerced to a raster object.
  }
    \item{\code{width}}{A numeric vector or unit object specifying width.}
    \item{\code{height}}{A numeric vector or unit object specifying height.}
    \item{\code{just}}{The justification of the rectangle
    relative to its (x, y) location.  If there are two values, the first
    value specifies horizontal justification and the second value specifies
    vertical justification.  Possible string values are: \code{"left"},
    \code{"right"}, \code{"centre"}, \code{"center"}, \code{"bottom"},
    and \code{"top"}.  For numeric values, 0 means left alignment
    and 1 means right alignment.
  }
    \item{\code{hjust}}{A numeric vector specifying horizontal justification.
    If specified, overrides the \code{just} setting.}
    \item{\code{vjust}}{A numeric vector specifying vertical justification.
    If specified, overrides the \code{just} setting.}
    \item{\code{name}}{ A character identifier. }
    \item{\code{gp}}{An object of class \code{"gpar"}, typically the output
    from a call to the function \code{\link[grid]{gpar}}.  This is basically
    a list of graphical parameter settings.}
    \item{\code{vp}}{A Grid viewport object (or NULL).}
    \item{\code{interpolate}}{
    A logical value indicating whether to linearly interpolate the
    image (the alternative is to use nearest-neighbour interpolation,
    which gives a more blocky result).
  }
  }}

\item{default.units}{A string indicating the default units to use
    if \code{x}, \code{y}, \code{width}, or \code{height}
    are only given as numeric vectors.}
}
\value{
A rasterGrob object
}
\description{
This is a convenience function that helps in creating \link{rasterGrob} with the
correct settings for the glyph. It takes inot account the sizing and offset
returned by \code{\link[=glyph_raster]{glyph_raster()}} and allows you to only consider the baseline
position of the glyph.
}
\examples{
font <- font_info()
glyph <- glyph_info("R", path = font$path, index = font$index)

R <- glyph_raster(glyph$index, font$path, font$index, size = 150)

grob <- glyph_raster_grob(R[[1]], 50, 50)

grid::grid.newpage()
# Mark the baseline location
grid::grid.points(50, 50, default.units = "bigpts")
# Draw the glyph
grid::grid.draw(grob)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/match_font.R
\name{match_font}
\alias{match_font}
\title{Find a system font by name and style}
\usage{
match_font(family, italic = FALSE, bold = FALSE)
}
\arguments{
\item{family}{The name of the font family}

\item{italic, bold}{logicals indicating the font style}
}
\value{
A list containing the path locating the font file and the 0-based
index of the font in the file.
}
\description{
This function locates the font file (and index) best matching a name and
optional style (italic/bold). A font file will be returned even if a match
isn't found, but it is not necessarily similar to the requested family and
it should not be relied on for font substitution. The aliases \code{"sans"},
\code{"serif"}, and \code{"mono"} match to the system default sans-serif, serif, and
mono fonts respectively (\code{""} is equivalent to \code{"sans"}).
}
\examples{
# Get the system default sans-serif font in italic
match_font('sans', italic = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/match_fonts.R
\name{as_font_weight}
\alias{as_font_weight}
\alias{as_font_width}
\title{Convert weight and width to numerics}
\usage{
as_font_weight(weight)

as_font_width(width)
}
\arguments{
\item{weight, width}{character vectors with valid names for weight or width}
}
\value{
An integer vector matching the length of the input
}
\description{
It is often more natural to describe font weight and width with names rather
than numbers (e.g. "bold" or "condensed"), but underneath these names are
matched to numeric values. These two functions are used to retrieve the
numeric counterparts to names
}
\examples{
as_font_weight(
  c("undefined", "thin", "ultralight", "light", "normal", "medium", "semibold",
    "bold", "ultrabold", "heavy")
)

as_font_width(
  c("undefined", "ultracondensed", "extracondensed", "condensed", "semicondensed",
  "normal", "semiexpanded", "expanded", "extraexpanded", "ultraexpanded")
)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system_fonts.R
\name{reset_font_cache}
\alias{reset_font_cache}
\title{Reset the system font cache}
\usage{
reset_font_cache()
}
\description{
Building the list of system fonts is time consuming and is therefore cached.
This, in turn, means that changes to the system fonts (i.e. installing new
fonts), will not propagate to systemfonts. The solution is to reset the
cache, which will result in the next call to e.g. \code{\link[=match_fonts]{match_fonts()}} will
trigger a rebuild of the cache.
}
\examples{
all_fonts <- system_fonts()

##-- Install a new font on the system --##

all_fonts_new <- system_fonts()

## all_fonts_new will be equal to all_fonts

reset_font_cache()

all_fonts_new <- system_fonts()

## all_fonts_new will now contain the new font

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_variation.R
\name{font_variation}
\alias{font_variation}
\title{Define axis coordinates for variable fonts}
\usage{
font_variation(
  italic = NULL,
  weight = NULL,
  width = NULL,
  slant = NULL,
  optical_sizing = NULL,
  ...
)
}
\arguments{
\item{italic}{Value between 0 and 1. Usually treated as a boolean rather than
a continuous axis}

\item{weight}{Usually a value between 100 and 900 though fonts can limit or
expand the supported range. Weight names are also allowed (see
\code{\link[=as_font_weight]{as_font_weight()}})}

\item{width}{Usually a value between 1 and 9 though fonts can limit or
expand the supported range. Width names are also allowed (see
\code{\link[=as_font_width]{as_font_width()}})}

\item{slant}{The angular slant of the font, usually between -90 and 90
(negative values tilt in the "standard" italic way)}

\item{optical_sizing}{Stroke thickness compensation for the glyphs. During
rendering the thickness of the stroke is usually increased when the font is
set at small sizes to increase readability. Set this to the size of the font
to get the "expected" look at that size.}

\item{...}{Further axes and coordinate settings}
}
\value{
A \code{font_variation} object with coordinates for the provided axes
}
\description{
Variable fonts is a technology that allows font designers to encode the full,
continuous font space of a typeface into a single font file and then have the
user set the coordinates of the variable axes to define the font. So, rather
than having a font file for bold, bold + italic, thin, and thin + italic,
etc. it is all encoded in a single file with a continuous range of all axes
(e.g. weight doesn't have to be one of the 9 standard weights but can be
anything in between). There are 5 standard axes that fonts can use, but font
designers are free to define their own completely arbitrary axes as well. You
can use \code{\link[=font_info]{font_info()}} to see which axes a font defines along with their value
range and default setting. Values given as \code{font_variation()} will always win
over the conventional setting \emph{if} the axis is present in the font. For
example, setting \code{weight = "bold"} along with
\code{variation = font_variation(weight = 650)} will eventually request a weight
of \code{650} (helfway between semibold and bold), assuming the weight-axis is
present in the font. For clarity however, it is advised that \code{font_variation()}
is only used for axes that can otherwise not be accessed by "top-level"
arguments.
}
\note{
systemfonts uses a scale of width values ranging from 1-9 while the
width axis uses a different scale (0.5 - 2.0) going from half as wide to
twice as wide as "normal". When using the \code{width} argument the coordinate
values is automatically converted. If you set the value based on the width
tag (\code{wdth}) then no conversion will happen.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev_strings.R
\name{string_widths_dev}
\alias{string_widths_dev}
\title{Get string widths as measured by the current device}
\usage{
string_widths_dev(
  strings,
  family = "",
  face = 1,
  size = 12,
  cex = 1,
  unit = "cm"
)
}
\arguments{
\item{strings}{A character vector of strings to measure}

\item{family}{The font families to use. Will get recycled}

\item{face}{The font faces to use. Will get recycled}

\item{size}{The font size to use. Will get recycled}

\item{cex}{The cex multiplier to use. Will get recycled}

\item{unit}{The unit to return the width in. Either \code{"cm"}, \code{"inches"},
\code{"device"}, or \code{"relative"}}
}
\value{
A numeric vector with the width of each of the strings given in
\code{strings} in the unit given in \code{unit}
}
\description{
For certain composition tasks it is beneficial to get the width of a string
as interpreted by the device that is going to plot it. grid provides this
through construction of a \code{textGrob} and then converting the corresponding
grob width to e.g. cm, but this comes with a huge overhead.
\code{string_widths_dev()} provides direct, vectorised, access to the graphic
device for as high performance as possible.
}
\examples{
# Get the widths as measured in cm (default)
string_widths_dev(c('a string', 'an even longer string'))

}
\seealso{
Other device metrics: 
\code{\link{string_metrics_dev}()}
}
\concept{device metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev_strings.R
\name{string_metrics_dev}
\alias{string_metrics_dev}
\title{Get string metrics as measured by the current device}
\usage{
string_metrics_dev(
  strings,
  family = "",
  face = 1,
  size = 12,
  cex = 1,
  unit = "cm"
)
}
\arguments{
\item{strings}{A character vector of strings to measure}

\item{family}{The font families to use. Will get recycled}

\item{face}{The font faces to use. Will get recycled}

\item{size}{The font size to use. Will get recycled}

\item{cex}{The cex multiplier to use. Will get recycled}

\item{unit}{The unit to return the width in. Either \code{"cm"}, \code{"inches"},
\code{"device"}, or \code{"relative"}}
}
\value{
A data.frame with \code{width}, \code{ascent}, and \code{descent} columns giving the
metrics in the requested unit.
}
\description{
This function is much like \code{\link[=string_widths_dev]{string_widths_dev()}} but also returns the ascent
and descent of the string making it possible to construct a tight bounding
box around the string.
}
\examples{
# Get the metrics as measured in cm (default)
string_metrics_dev(c('some text', 'a string with descenders'))

}
\seealso{
Other device metrics: 
\code{\link{string_widths_dev}()}
}
\concept{device metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/register_font.R
\name{register_font}
\alias{register_font}
\alias{registry_fonts}
\alias{clear_registry}
\title{Register font collections as families}
\usage{
register_font(
  name,
  plain,
  bold = plain,
  italic = plain,
  bolditalic = plain,
  features = font_feature()
)

registry_fonts()

clear_registry()
}
\arguments{
\item{name}{The name the collection will be known under (i.e. \emph{family})}

\item{plain, bold, italic, bolditalic}{Fontfiles for the different faces of the
collection. can either be a filepath or a list containing a filepath and an
index (only for font files containing multiple fonts). If not given it will
default to the \code{plain} specification.}

\item{features}{A \code{\link{font_feature}} object describing the specific OpenType
font features to turn on for the registered font.}
}
\value{
\code{register_font()} and \code{clear_registry()} returns \code{NULL} invisibly.
\code{registry_fonts()} returns a data table in the same style as \code{\link[=system_fonts]{system_fonts()}}
though less detailed and not based on information in the font file.
}
\description{
By design, systemfonts searches the fonts installed natively on the system.
It is possible, however, to register other fonts from e.g. font packages or
local font files, that will get searched before searching any installed
fonts. You can always get an overview over all registered fonts with the
\code{registry_fonts()} function that works as a registry focused analogue to
\code{\link[=system_fonts]{system_fonts()}}. If you wish to clear out the registry, you can either
restart the R session or call \code{clear_registry()}.
}
\details{
\code{register_font} also makes it possible to use system fonts with traits that
is not covered by the graphic engine in R. In plotting operations it is only
possible to specify a family name and whether or not the font should be bold
and/or italic. There are numerous fonts that will never get matched to this,
especially because bold is only one of many weights.

Apart from granting a way to use new varieties of fonts, font registration
also allows you to override the default \code{sans}, \code{serif}, and \code{mono} mappings,
simply by registering a collection to the relevant default name. As
registered fonts are searched first it will take precedence over the default.
}
\section{Font matching}{
During font matching, systemfonts has to look in three different locations.
The font registry (populated by \code{\link[=register_font]{register_font()}}/\code{\link[=register_variant]{register_variant()}}), the
local fonts (populated with \code{\link[=add_fonts]{add_fonts()}}/\code{\link[=scan_local_fonts]{scan_local_fonts()}}), and the
fonts installed on the system. It does so in that order: registry > local >
installed.

The matching performed at each step also differs. The fonts in the registry
is only matched by family name. The local fonts are matched based on all the
provided parameters (family, weight, italic, etc) in a way that is local to
systemfonts, but try to emulate the system native matching. The installed
fonts are matched using the system native matching functionality on macOS and
Linux. On Windows the installed fonts are read from the system registry and
matched using the same approach as for local fonts. Matching will always find
a font no matter what you throw at it, defaulting to "sans" if nothing else
is found.
}

\examples{
\dontshow{if (nrow(system_fonts()) >= 4) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Create a random font collection
fonts <- system_fonts()
random_fonts <- sample(nrow(fonts), 4)
register_font(
  'random',
  plain = list(fonts$path[random_fonts[1]], fonts$index[random_fonts[1]]),
  bold = list(fonts$path[random_fonts[2]], fonts$index[random_fonts[2]]),
  italic = list(fonts$path[random_fonts[3]], fonts$index[random_fonts[3]]),
  bolditalic = list(fonts$path[random_fonts[4]], fonts$index[random_fonts[4]])
)

# Look at your creation
registry_fonts()

# Reset
clear_registry()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/web_fonts.R
\name{fonts_as_import}
\alias{fonts_as_import}
\title{Create import specifications for web content}
\usage{
fonts_as_import(
  family,
  italic = NULL,
  weight = NULL,
  width = NULL,
  ...,
  type = c("url", "import", "link"),
  may_embed = TRUE,
  repositories = c("Bunny Fonts", "Google Fonts", "Font Library")
)
}
\arguments{
\item{family}{The name of the font families to match}

\item{italic}{logical indicating the font slant}

\item{weight}{The weight to query for, either in numbers (\code{0}, \code{100}, \code{200},
\code{300}, \code{400}, \code{500}, \code{600}, \code{700}, \code{800}, or \code{900}) or strings (\code{"undefined"},
\code{"thin"}, \code{"ultralight"}, \code{"light"}, \code{"normal"}, \code{"medium"}, \code{"semibold"},
\code{"bold"}, \code{"ultrabold"}, or \code{"heavy"}). \code{NA} will be interpreted as
\code{"undefined"}/\code{0}}

\item{width}{The width to query for either in numbers (\code{0}, \code{1}, \code{2},
\code{3}, \code{4}, \code{5}, \code{6}, \code{7}, \code{8}, or \code{9}) or strings (\code{"undefined"},
\code{"ultracondensed"}, \code{"extracondensed"}, \code{"condensed"}, \code{"semicondensed"},
\code{"normal"}, \code{"semiexpanded"}, \code{"expanded"}, \code{"extraexpanded"}, or
\code{"ultraexpanded"}). \code{NA} will be interpreted as \code{"undefined"}/\code{0}}

\item{...}{Additional arguments passed on to the specific functions for the
repositories. Currently:
\itemize{
\item \strong{Google Fonts and Bunny Fonts:}
\itemize{
\item \code{text} A piece of text containing the glyphs required. Using this can
severely cut down on the size of the required download
\item \code{display} One of \code{"auto"}, \code{"block"}, \code{"swap"}, \code{"fallback"}, or
\code{"optional"}. Controls how the text is displayed while the font is
downloading.
}
}}

\item{type}{The type of return value. \code{"url"} returns the bare url pointing
to the style sheet. \code{"import"} returns the stylesheet as an import statement
(\verb{@import url(<url>)}). \code{"link"} returns the stylesheet as a link tag
(\verb{<link rel="stylesheet" href="<url>"/>})}

\item{may_embed}{Logical. Should fonts that can't be found in the provided
repositories be embedded as data-URLs. This is only possible if the font is
available locally and in a \code{woff2}, \code{woff}, \code{otf}, or \code{ttf} file.}

\item{repositories}{The repositories to try looking for the font. Currently
\code{"Bunny Fonts"}, \code{"Google Fonts"}, and \code{"Font Library"} are supported. Set
this to \code{NULL} together with \code{may_embed = TRUE} to force embedding of the
font data.}
}
\value{
A character vector with stylesheet specifications according to \code{type}
}
\description{
If you create content in a text-based format such as HTML or SVG you need to
make sure that the font is available on the computer where it is viewed. This
can be achieved through the use of stylesheets that can either be added with
a \verb{<link>} tag or inserted with an \verb{@import} statement. This function
facilitates the creation of either of these (or the bare URL to the
stylesheet). It can rely on the Bunny Fonts, Google Fonts and/or Font Library
repositories for serving the fonts. If the requested font is not found it can
optionally hard code the data into the stylesheet.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape_string.R
\name{shape_string}
\alias{shape_string}
\title{Calculate glyph positions for strings}
\usage{
shape_string(
  strings,
  id = NULL,
  family = "",
  italic = FALSE,
  bold = FALSE,
  size = 12,
  res = 72,
  lineheight = 1,
  align = "left",
  hjust = 0,
  vjust = 0,
  width = NA,
  tracking = 0,
  indent = 0,
  hanging = 0,
  space_before = 0,
  space_after = 0,
  path = NULL,
  index = 0
)
}
\arguments{
\item{strings}{A character vector of strings to shape}

\item{id}{A vector grouping the strings together. If strings share an id the
shaping will continue between strings}

\item{family}{The name of the font family}

\item{italic}{logicals indicating the font style}

\item{bold}{logicals indicating the font style}

\item{size}{The pointsize of the font to use for size related measures}

\item{res}{The ppi of the size related mesures}

\item{lineheight}{A multiplier for the lineheight}

\item{align}{Within text box alignment, either \code{'left'}, \code{'center'}, or
\code{'right'}}

\item{hjust, vjust}{The justification of the textbox surrounding the text}

\item{width}{The requested with of the string in inches. Setting this to
something other than \code{NA} will turn on word wrapping.}

\item{tracking}{Tracking of the glyphs (space adjustment) measured in 1/1000
em.}

\item{indent}{The indent of the first line in a paragraph measured in inches.}

\item{hanging}{The indent of the remaining lines in a paragraph measured in
inches.}

\item{space_before, space_after}{The spacing above and below a paragraph,
measured in points}

\item{path, index}{path an index of a font file to circumvent lookup based on
family and style}
}
\value{
A list with two element: \code{shape} contains the position of each glyph,
relative to the origin in the enclosing textbox. \code{metrics} contain metrics
about the full strings.

\code{shape} is a data.frame with the following columns:
\describe{
\item{glyph}{The glyph as a character}
\item{index}{The index of the glyph in the font file}
\item{metric_id}{The index of the string the glyph is part of (referencing a row in the \code{metrics} data.frame)}
\item{string_id}{The index of the string the glyph came from (referencing an element in the \code{strings} input)}
\item{x_offset}{The x offset in pixels from the origin of the textbox}
\item{y_offset}{The y offset in pixels from the origin of the textbox}
\item{x_mid}{The x offset in pixels to the middle of the glyph, measured from the origin of the glyph}
}

\code{metrics} is a data.frame with the following columns:
\describe{
\item{string}{The text the string consist of}
\item{width}{The width of the string}
\item{height}{The height of the string}
\item{left_bearing}{The distance from the left edge of the textbox and the leftmost glyph}
\item{right_bearing}{The distance from the right edge of the textbox and the rightmost glyph}
\item{top_bearing}{The distance from the top edge of the textbox and the topmost glyph}
\item{bottom_bearing}{The distance from the bottom edge of the textbox and the bottommost glyph}
\item{left_border}{The position of the leftmost edge of the textbox related to the origin}
\item{top_border}{The position of the topmost edge of the textbox related to the origin}
\item{pen_x}{The horizontal position of the next glyph after the string}
\item{pen_y}{The vertical position of the next glyph after the string}
}
}
\description{
Do basic text shaping of strings. This function will use freetype to
calculate advances, doing kerning if possible. It will not perform any font
substitution or ligature resolving and will thus be much in line with how
the standard graphic devices does text shaping. Inputs are recycled to the
length of \code{strings}.
}
\examples{
string <- "This is a long string\nLook; It spans multiple lines\nand all"

# Shape with default settings
shape_string(string)

# Mix styles within the same string
string <- c(
  "This string will have\na ",
  "very large",
  " text style\nin the middle"
)

shape_string(string, id = c(1, 1, 1), size = c(12, 24, 12))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_info.R
\name{glyph_info}
\alias{glyph_info}
\title{Query glyph-specific information from fonts}
\usage{
glyph_info(
  glyphs,
  family = "",
  italic = FALSE,
  weight = "normal",
  width = "undefined",
  size = 12,
  res = 72,
  path = NULL,
  index = 0,
  variation = font_variation(),
  bold = deprecated()
)
}
\arguments{
\item{glyphs}{A vector of glyphs. Strings will be split into separate glyphs
automatically}

\item{family}{The name of the font families to match}

\item{italic}{logical indicating the font slant}

\item{weight}{The weight to query for, either in numbers (\code{0}, \code{100}, \code{200},
\code{300}, \code{400}, \code{500}, \code{600}, \code{700}, \code{800}, or \code{900}) or strings (\code{"undefined"},
\code{"thin"}, \code{"ultralight"}, \code{"light"}, \code{"normal"}, \code{"medium"}, \code{"semibold"},
\code{"bold"}, \code{"ultrabold"}, or \code{"heavy"}). \code{NA} will be interpreted as
\code{"undefined"}/\code{0}}

\item{width}{The width to query for either in numbers (\code{0}, \code{1}, \code{2},
\code{3}, \code{4}, \code{5}, \code{6}, \code{7}, \code{8}, or \code{9}) or strings (\code{"undefined"},
\code{"ultracondensed"}, \code{"extracondensed"}, \code{"condensed"}, \code{"semicondensed"},
\code{"normal"}, \code{"semiexpanded"}, \code{"expanded"}, \code{"extraexpanded"}, or
\code{"ultraexpanded"}). \code{NA} will be interpreted as \code{"undefined"}/\code{0}}

\item{size}{The pointsize of the font to use for size related measures}

\item{res}{The ppi of the size related measures}

\item{path, index}{path an index of a font file to circumvent lookup based on
family and style}

\item{variation}{A \code{font_variation} object or a list of them to control
variable fonts}

\item{bold}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{weight = "bold"} instead}
}
\value{
A data.frame with information about each glyph, containing the following
columns:

\describe{
\item{glyph}{The glyph as a character}
\item{index}{The index of the glyph in the font file}
\item{width}{The width of the glyph}
\item{height}{The height of the glyph}
\item{x_bearing}{The horizontal distance from the origin to the leftmost part of the glyph}
\item{y_bearing}{The vertical distance from the origin to the top part of the glyph}
\item{x_advance}{The horizontal distance to move the cursor after adding the glyph}
\item{y_advance}{The vertical distance to move the cursor after adding the glyph}
\item{bbox}{The tight bounding box surrounding the glyph}
}
}
\description{
This function allows you to extract information about the individual glyphs
in a font, based on a specified size. All size related measures are in
pixel-units. The function is vectorised to the length of the \code{glyphs} vector.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_outline.R
\name{plot_glyph_stats}
\alias{plot_glyph_stats}
\title{Create a visual representation of what the various glyph stats mean}
\usage{
plot_glyph_stats(
  glyph,
  family = "",
  italic = FALSE,
  weight = "normal",
  width = "undefined",
  size = 12,
  res = 72,
  variation = font_variation(),
  path = NULL,
  index = 0
)
}
\arguments{
\item{glyph}{The character to plot}

\item{family}{The name of the font families to match}

\item{italic}{logical indicating the font slant}

\item{weight}{The weight to query for, either in numbers (\code{0}, \code{100}, \code{200},
\code{300}, \code{400}, \code{500}, \code{600}, \code{700}, \code{800}, or \code{900}) or strings (\code{"undefined"},
\code{"thin"}, \code{"ultralight"}, \code{"light"}, \code{"normal"}, \code{"medium"}, \code{"semibold"},
\code{"bold"}, \code{"ultrabold"}, or \code{"heavy"}). \code{NA} will be interpreted as
\code{"undefined"}/\code{0}}

\item{width}{The width to query for either in numbers (\code{0}, \code{1}, \code{2},
\code{3}, \code{4}, \code{5}, \code{6}, \code{7}, \code{8}, or \code{9}) or strings (\code{"undefined"},
\code{"ultracondensed"}, \code{"extracondensed"}, \code{"condensed"}, \code{"semicondensed"},
\code{"normal"}, \code{"semiexpanded"}, \code{"expanded"}, \code{"extraexpanded"}, or
\code{"ultraexpanded"}). \code{NA} will be interpreted as \code{"undefined"}/\code{0}}

\item{size}{The pointsize of the font to use for size related measures}

\item{res}{The ppi of the size related measures}

\item{variation}{A \code{font_variation} object or a list of them to control
variable fonts}

\item{path, index}{path an index of a font file to circumvent lookup based on
family and style}
}
\value{
This function is called for its side effects
}
\description{
This function helps you understand the concepts of width, height, bearing,
and advance by annotating a glyph with the various measures
}
\examples{
plot_glyph_stats("g")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/register_font.R
\name{add_fonts}
\alias{add_fonts}
\alias{scan_local_fonts}
\alias{clear_local_fonts}
\title{Add local font files to the search path}
\usage{
add_fonts(files)

scan_local_fonts()

clear_local_fonts()
}
\arguments{
\item{files}{A character vector of font file paths or urls to add}
}
\value{
This function is called for its sideeffects
}
\description{
systemfonts is mainly about getting system native access to the fonts
installed on the OS you are executing the code on. However, you may want to
access fonts without doing a full installation, either because you want your
project to be reproducible on all systems, because you don't have
administrator privileges on the system, or for a different reason entirely.
\code{add_fonts()} provide a way to side load font files so that they are found
during font matching. The function differs from \code{\link[=register_font]{register_font()}} and
\code{\link[=register_variant]{register_variant()}} in that they add the font file as-is using the family
name etc that are provided by the font. \code{scan_local_fonts()} is run when
systemfonts is loaded and will automatically add font files stored in
\code{./fonts} (project local) and \verb{~/fonts} (user local).
}
\section{Font matching}{
During font matching, systemfonts has to look in three different locations.
The font registry (populated by \code{\link[=register_font]{register_font()}}/\code{\link[=register_variant]{register_variant()}}), the
local fonts (populated with \code{\link[=add_fonts]{add_fonts()}}/\code{\link[=scan_local_fonts]{scan_local_fonts()}}), and the
fonts installed on the system. It does so in that order: registry > local >
installed.

The matching performed at each step also differs. The fonts in the registry
is only matched by family name. The local fonts are matched based on all the
provided parameters (family, weight, italic, etc) in a way that is local to
systemfonts, but try to emulate the system native matching. The installed
fonts are matched using the system native matching functionality on macOS and
Linux. On Windows the installed fonts are read from the system registry and
matched using the same approach as for local fonts. Matching will always find
a font no matter what you throw at it, defaulting to "sans" if nothing else
is found.
}

\examples{
# example code
empty_font <- system.file("unfont.ttf", package = "systemfonts")

add_fonts(empty_font)

clear_local_fonts()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_outline.R
\name{glyph_outline}
\alias{glyph_outline}
\title{Get the outline of glyphs}
\usage{
glyph_outline(
  glyph,
  path,
  index = 0,
  size = 12,
  tolerance = 0.2,
  variation = font_variation(),
  verbose = FALSE
)
}
\arguments{
\item{glyph}{The index of the glyph in the font file}

\item{path}{The path to the font file encoding the glyph}

\item{index}{The index of the font in the font file}

\item{size}{The size of the font in big points (1/72 inch)}

\item{tolerance}{The deviation tolerance for decomposing bezier curves of the
glyph. Given in the same unit as size. Smaller values give more detailed
polygons}

\item{variation}{A \code{font_variation} object or a list of them to control
variable fonts}

\item{verbose}{Should font and glyph loading errors be reported as warnings}
}
\value{
A data frame giving the outlines of the glyphs provide in \code{glyph}. It
contains the columns \code{glyph} pointing to the element in the input it relates
to, \code{contour} enumerating the contours the glyph consists of, and \code{x} and \code{y}
giving the coordinates in big points
}
\description{
This function allows you to retrieve the outline of glyphs as polygon
coordinates. The glyphs are given as indexes into a font file and not as
characters allowing you to retrieve outlines for glyphs that doesn't have a
character counterpoint. Glyphs that are given as bitmaps are ignored.
}
\examples{
# Get the shape of s in the default font
font <- font_info()
glyph <- glyph_info("s", path = font$path, index = font$index)

s <- glyph_outline(glyph$index, font$path, font$index, size = 150)

plot(s$x, s$y, type = 'l')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_fallback.R
\name{font_fallback}
\alias{font_fallback}
\title{Get the fallback font for a given string}
\usage{
font_fallback(
  string,
  family = "",
  italic = FALSE,
  weight = "normal",
  width = "undefined",
  path = NULL,
  index = 0,
  variation = font_variation(),
  bold = deprecated()
)
}
\arguments{
\item{string}{The strings to find fallbacks for}

\item{family}{The name of the font families to match}

\item{italic}{logical indicating the font slant}

\item{weight}{The weight to query for, either in numbers (\code{0}, \code{100}, \code{200},
\code{300}, \code{400}, \code{500}, \code{600}, \code{700}, \code{800}, or \code{900}) or strings (\code{"undefined"},
\code{"thin"}, \code{"ultralight"}, \code{"light"}, \code{"normal"}, \code{"medium"}, \code{"semibold"},
\code{"bold"}, \code{"ultrabold"}, or \code{"heavy"}). \code{NA} will be interpreted as
\code{"undefined"}/\code{0}}

\item{width}{The width to query for either in numbers (\code{0}, \code{1}, \code{2},
\code{3}, \code{4}, \code{5}, \code{6}, \code{7}, \code{8}, or \code{9}) or strings (\code{"undefined"},
\code{"ultracondensed"}, \code{"extracondensed"}, \code{"condensed"}, \code{"semicondensed"},
\code{"normal"}, \code{"semiexpanded"}, \code{"expanded"}, \code{"extraexpanded"}, or
\code{"ultraexpanded"}). \code{NA} will be interpreted as \code{"undefined"}/\code{0}}

\item{path, index}{path and index of a font file to circumvent lookup based on
family and style}

\item{variation}{A \code{font_variation} object or a list of them to control
variable fonts}

\item{bold}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{weight = "bold"} instead}
}
\value{
A data frame with a \code{path} and \code{index} column giving fallback for the
specified string and font combinations
}
\description{
A fallback font is a font to use as a substitute if the chosen font does not
contain the requested characters. Using font fallbacks means that the user
doesn't have to worry about mixing characters from different scripts or
mixing text and emojies. Fallback is calculated for the full string and the
result is platform specific. If no font covers all the characters in the
string an undefined "best match" is returned. The best approach is to figure
out which characters are not covered by your chosen font and figure out
fallbacks for these, rather than just request a fallback for the full string.
}
\examples{
font_fallback("\U0001f604") # Smile emoji

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_outline.R
\name{glyph_raster}
\alias{glyph_raster}
\title{Render glyphs to raster image}
\usage{
glyph_raster(
  glyph,
  path,
  index = 0,
  size = 12,
  res = 300,
  variation = font_variation(),
  col = "black",
  verbose = FALSE
)
}
\arguments{
\item{glyph}{The index of the glyph in the font file}

\item{path}{The path to the font file encoding the glyph}

\item{index}{The index of the font in the font file}

\item{size}{The size of the font in big points (1/72 inch)}

\item{res}{The resolution to render the glyphs to}

\item{variation}{A \code{font_variation} object or a list of them to control
variable fonts}

\item{col}{The color of the glyph assuming the glyph doesn't have a native
coloring}

\item{verbose}{Should font and glyph loading errors be reported as warnings}
}
\value{
A list of nativeRaster objects (or \code{NULL} if it failed to render a
given glyph). The nativeRasters have additional attributes attached. \code{"size"}
will give the size of the glyph in big points and \code{"offset"} will give the
location of the top-left corner of the raster with respect to where it should
be rendered.
}
\description{
Not all glyphs are encoded as vector outlines (emojis often not). Even for
fonts that provide an outline you might be interested in a raster version.
This function gives you just that. It converts a glyph into an optimized
raster object that can be plotted with e.g. \code{\link[graphics:rasterImage]{graphics::rasterImage()}} or
\code{\link[grid:grid.raster]{grid::grid.raster()}}. For convenience, you can also use
\code{\link[=glyph_raster_grob]{glyph_raster_grob()}} for plotting the result.
}
\examples{
font <- font_info()
glyph <- glyph_info("R", path = font$path, index = font$index)

R <- glyph_raster(glyph$index, font$path, font$index, size = 150)

plot.new()
plot.window(c(0,150), c(0, 150), asp = 1)
rasterImage(R[[1]], 0, 0, attr(R[[1]], "size")[2], attr(R[[1]], "size")[1])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape_string.R
\name{string_width}
\alias{string_width}
\title{Calculate the width of a string, ignoring new-lines}
\usage{
string_width(
  strings,
  family = "",
  italic = FALSE,
  weight = "normal",
  width = "undefined",
  size = 12,
  res = 72,
  include_bearing = TRUE,
  path = NULL,
  index = 0,
  bold = deprecated()
)
}
\arguments{
\item{strings}{A character vector of strings}

\item{family}{The name of the font families to match}

\item{italic}{logical indicating the font slant}

\item{weight}{The weight to query for, either in numbers (\code{0}, \code{100}, \code{200},
\code{300}, \code{400}, \code{500}, \code{600}, \code{700}, \code{800}, or \code{900}) or strings (\code{"undefined"},
\code{"thin"}, \code{"ultralight"}, \code{"light"}, \code{"normal"}, \code{"medium"}, \code{"semibold"},
\code{"bold"}, \code{"ultrabold"}, or \code{"heavy"}). \code{NA} will be interpreted as
\code{"undefined"}/\code{0}}

\item{width}{The width to query for either in numbers (\code{0}, \code{1}, \code{2},
\code{3}, \code{4}, \code{5}, \code{6}, \code{7}, \code{8}, or \code{9}) or strings (\code{"undefined"},
\code{"ultracondensed"}, \code{"extracondensed"}, \code{"condensed"}, \code{"semicondensed"},
\code{"normal"}, \code{"semiexpanded"}, \code{"expanded"}, \code{"extraexpanded"}, or
\code{"ultraexpanded"}). \code{NA} will be interpreted as \code{"undefined"}/\code{0}}

\item{size}{The pointsize of the font to use for size related measures}

\item{res}{The ppi of the size related measures}

\item{include_bearing}{Logical, should left and right bearing be included in
the string width?}

\item{path, index}{path and index of a font file to circumvent lookup based on
family and style}

\item{bold}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{weight = "bold"} instead}
}
\value{
A numeric vector giving the width of the strings in pixels. Use the
provided \code{res} value to convert it into absolute values.
}
\description{
This is a very simple alternative to \code{\link[=shape_string]{shape_string()}} that simply calculates
the width of strings without taking any newline into account. As such it is
suitable to calculate the width of words or lines that has already been
splitted by \verb{\\n}. Input is recycled to the length of \code{strings}.
}
\examples{
strings <- c('A short string', 'A very very looong string')
string_width(strings)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/match_fonts.R
\name{match_fonts}
\alias{match_fonts}
\alias{match_font}
\title{Find a system font by name and style}
\usage{
match_fonts(family, italic = FALSE, weight = "normal", width = "undefined")

match_font(family, italic = FALSE, bold = FALSE)
}
\arguments{
\item{family}{The name of the font families to match}

\item{italic}{logical indicating the font slant}

\item{weight}{The weight to query for, either in numbers (\code{0}, \code{100}, \code{200},
\code{300}, \code{400}, \code{500}, \code{600}, \code{700}, \code{800}, or \code{900}) or strings (\code{"undefined"},
\code{"thin"}, \code{"ultralight"}, \code{"light"}, \code{"normal"}, \code{"medium"}, \code{"semibold"},
\code{"bold"}, \code{"ultrabold"}, or \code{"heavy"}). \code{NA} will be interpreted as
\code{"undefined"}/\code{0}}

\item{width}{The width to query for either in numbers (\code{0}, \code{1}, \code{2},
\code{3}, \code{4}, \code{5}, \code{6}, \code{7}, \code{8}, or \code{9}) or strings (\code{"undefined"},
\code{"ultracondensed"}, \code{"extracondensed"}, \code{"condensed"}, \code{"semicondensed"},
\code{"normal"}, \code{"semiexpanded"}, \code{"expanded"}, \code{"extraexpanded"}, or
\code{"ultraexpanded"}). \code{NA} will be interpreted as \code{"undefined"}/\code{0}}

\item{bold}{logical indicating whether the font weight}
}
\value{
A list containing the paths locating the font files, the 0-based
index of the font in the files and the features for the font in case a
registered font was located.
}
\description{
This function locates the font file (and index) best matching a name and
optional style. A font file will be returned even if a perfect match
isn't found, but it is not necessarily similar to the requested family and
it should not be relied on for font substitution. The aliases \code{"sans"},
\code{"serif"}, \code{"mono"}, \code{"symbol"}, and \code{"emoji"} match to their respective
system defaults (\code{""} is equivalent to \code{"sans"}). \code{match_font()} has been
deprecated in favour of \code{match_fonts()} which provides vectorisation, as well
as querying for different weights (rather than just "normal" and "bold") as
well as different widths.
}
\section{Font matching}{
During font matching, systemfonts has to look in three different locations.
The font registry (populated by \code{\link[=register_font]{register_font()}}/\code{\link[=register_variant]{register_variant()}}), the
local fonts (populated with \code{\link[=add_fonts]{add_fonts()}}/\code{\link[=scan_local_fonts]{scan_local_fonts()}}), and the
fonts installed on the system. It does so in that order: registry > local >
installed.

The matching performed at each step also differs. The fonts in the registry
is only matched by family name. The local fonts are matched based on all the
provided parameters (family, weight, italic, etc) in a way that is local to
systemfonts, but try to emulate the system native matching. The installed
fonts are matched using the system native matching functionality on macOS and
Linux. On Windows the installed fonts are read from the system registry and
matched using the same approach as for local fonts. Matching will always find
a font no matter what you throw at it, defaulting to "sans" if nothing else
is found.
}

\examples{
# Get the system default sans-serif font in italic
match_fonts('sans', italic = TRUE)

# Try to match it to a thin variant
match_fonts(c('sans', 'serif'), weight = "thin")

}
